import os
from datetime import datetime
from typing import List, Optional

import click

from vessl.openapi_client import ModelRepositoryCreateAPIInput, ResponseSimpleModelInfo
from vessl import list_experiment_output_files, list_experiments, vessl_api
from vessl.cli._base import (
    VesslGroup,
    vessl_argument,
    vessl_conditional_option,
    vessl_option,
)
from vessl.cli._util import (
    print_data,
  print_table,
    print_volume_files,
  truncate_datetime,
)
from vessl.util.fmt import format_size, format_url
from vessl.util.prompt import prompt_confirm, prompt_text, prompt_choices, prompt_checkbox, generic_prompter
from vessl.util.echo import print_info, print_success, print_error
from vessl.util.endpoint import Endpoint
from vessl.cli.dataset import download_dest_path_prompter
from vessl.cli.organization import organization_name_option
from vessl.cli.project import project_name_option
from vessl.experiment import read_experiment_by_id
from vessl.models.vessl_model import VesslModel
from vessl.kernel_image import get_recent_framework_and_matchin_cuda_versions
from vessl.model import (
    _make_endpoint_file_for_hf,
    create_model,
    create_model_repository,
    delete_model_volume_file,
    download_model_volume_file,
    list_model_repositories,
    list_model_volume_files,
    list_models,
    read_model,
    read_model_repository,
    upload_model_volume_file,
)
from vessl.organization import _get_organization_name
from vessl.service import serve_model
from vessl.util.constant import MODEL_SOURCE_EXPERIMENT, MODEL_SOURCE_LOCAL


def model_repository_status(latest_model: ResponseSimpleModelInfo) -> str:
    if latest_model:
        return latest_model.status
    return "Empty repository"


def model_repository_update_dt(
    latest_model: ResponseSimpleModelInfo, default: datetime
) -> datetime:
    if latest_model:
        return truncate_datetime(latest_model.created_dt)
    return truncate_datetime(default)


def model_repository_name_callback(ctx: click.Context, param: click.Parameter, value: str) -> str:
    if value:
        ctx.obj["repository"] = value
    return value


def model_repository_name_prompter(
    ctx: click.Context,
    param: click.Parameter,
    value: str,
) -> str:
    model_repositories = list_model_repositories()
    if len(model_repositories) == 0:
        raise click.UsageError(
            message="Create model repository with `vessl model-repository create`"
        )
    repository = prompt_choices("Model repository", [x.name for x in model_repositories])
    ctx.obj["repository"] = repository
    return repository


@click.command(name="model-repository", cls=VesslGroup)
def model_repository_cli():
    pass


@model_repository_cli.vessl_command()
@vessl_argument("name", type=click.STRING, required=True, prompter=model_repository_name_prompter)
@organization_name_option
def read(name: str):
    model_repository = read_model_repository(repository_name=name)
    print_data(
        {
            "ID": model_repository.id,
            "Name": model_repository.name,
            "Description": model_repository.description,
            "Status": model_repository_status(model_repository.model_summary.latest_model),
            "Organization": model_repository.organization.name,
            "Created": truncate_datetime(model_repository.created_dt),
            "Updated": truncate_datetime(model_repository.updated_dt),
        }
    )
    print_info(
        f"For more info: {format_url(Endpoint.model_repository.format(model_repository.organization.name, model_repository.name))}"
    )


@model_repository_cli.vessl_command()
@organization_name_option
@vessl_option(
    "--limit",
    type=click.INT,
    required=False,
    default=None,
)
def list(limit: Optional[int]):
    model_repositories = list_model_repositories(limit=limit)
    print_table(
        model_repositories,
        ["Name", "Status", "Models", "Created", "Updated"],
        lambda x: [
            x.name,
            model_repository_status(x.model_summary.latest_model),
            x.model_summary.total,
            truncate_datetime(x.created_dt),
            model_repository_update_dt(x.model_summary.latest_model, x.updated_dt),
        ],
    )


@model_repository_cli.vessl_command()
@vessl_argument(
    "name",
    type=click.STRING,
    required=True,
    prompter=generic_prompter("Model repository name"),
)
@click.option("-m", "--description", type=click.STRING)
@organization_name_option
def create(
    name: str,
    description: str,
):
    model_repository = create_model_repository(
        name=name,
        description=description,
    )
    print_success(
        f"Created '{model_repository.name}'.\n"
        f"For more info: {format_url(Endpoint.model_repository.format(model_repository.organization.name, model_repository.name))}"
    )


def model_number_prompter(
    ctx: click.Context,
    param: click.Parameter,
    value: int,
) -> int:
    repository = ctx.obj.get("repository")
    if repository is None:
        raise click.BadArgumentUsage(
            message="Argument `REPOSITORY_NAME` must be specified before `MODEL_NUMBER`.",
        )
    models = list_models(repository_name=repository)
    return prompt_choices("Model", [x.number for x in models])


@click.command(name="model", cls=VesslGroup)
def model_cli():
    pass


@model_cli.vessl_command()
@vessl_argument(
    "repository_name",
    type=click.STRING,
    required=True,
    prompter=model_repository_name_prompter,
    callback=model_repository_name_callback,
)
@vessl_argument(
    "model_number",
    type=click.INT,
    required=True,
    prompter=model_number_prompter,
)
@click.option(
    "--install-reqs",
    type=click.BOOL,
    is_flag=True,
    help="Install requirements before serving.",
)
@click.option(
    "--remote",
    type=click.BOOL,
    is_flag=True,
    hidden=True,
)
@organization_name_option
def serve(repository_name: str, model_number: int, install_reqs: bool, remote: bool):
    serve_model(repository_name, model_number, install_reqs, remote)


@model_cli.vessl_command()
@vessl_argument(
    "repository_name",
    type=click.STRING,
    required=True,
    prompter=model_repository_name_prompter,
    callback=model_repository_name_callback,
)
@vessl_argument(
    "model_number",
    type=click.INT,
    required=True,
    prompter=model_number_prompter,
)
@organization_name_option
def read(repository_name: str, model_number: int):
    model = read_model(repository_name=repository_name, model_number=model_number)
    experiment_numbers = []
    if model.experiment:
        experiment_numbers.append(model.experiment.number)

    metrics_summary = "None"
    if model.metrics_summary:
        metrics_keys = model.metrics_summary.latest.keys()
        metrics_summary = {}
        for key in metrics_keys:
            metrics_summary[key] = model.metrics_summary.latest[key].value

    print_data(
        {
            "ID": model.id,
            "Number": model.number,
            "Name": model.name,
            "Status": model.status,
            "Related experiments": experiment_numbers,
            "Creator": model.created_by.username,
            "Created": truncate_datetime(model.created_dt),
            "Metrics summary": metrics_summary,
        }
    )
    print_info(
        f"For more info: {format_url(Endpoint.model.format(model.model_repository.organization.name, model.model_repository.name, model.number))}"
    )


@model_cli.vessl_command()
@vessl_argument(
    "repository",
    type=click.STRING,
    required=True,
    prompter=model_repository_name_prompter,
)
@organization_name_option
def list(repository: str):
    models = list_models(repository_name=repository)
    print_table(
        models,
        ["Number", "Created", "Status"],
        lambda x: [x.number, truncate_datetime(x.created_dt), x.status],
    )


@model_cli.vessl_command()
@vessl_argument(
    "repository_name",
    type=click.STRING,
    required=True,
    prompter=model_repository_name_prompter,
)
@vessl_argument(
    "model_number",
    type=click.INT,
    required=True,
    prompter=model_number_prompter,
)
@click.option("-p", "--path", type=click.Path(), default="", help="Defaults to root.")
@click.option("-r", "--recursive", is_flag=True)
@organization_name_option
def list_files(repository_name: str, model_number: int, path: str, recursive: bool):
    files = list_model_volume_files(
        repository_name=repository_name,
        model_number=model_number,
        need_download_url=False,
        path=path,
        recursive=recursive,
    )
    print_volume_files(files)


def model_source_prompter(
    ctx: click.Context,
    param: click.Parameter,
    value: str,
) -> str:
    return prompt_choices(
        "Source",
        [
            ("From an experiment", MODEL_SOURCE_EXPERIMENT),
            ("From local files", MODEL_SOURCE_LOCAL),
        ],
    )


def model_source_callback(
    ctx: click.Context,
    param: click.Parameter,
    value: str,
):
    if value:
        ctx.obj["model_source"] = value
    return value


def experiment_id_prompter(
    ctx: click.Context,
    param: click.Parameter,
    value: int,
) -> int:
    model_source = ctx.obj.get("model_source")
    if model_source is None:
        raise click.BadOptionUsage(
            option_name="--source",
            message="Model source (`--source`) must be specified before experiment_id (`--experiment-id`).",
        )
    if model_source == MODEL_SOURCE_EXPERIMENT:
        experiments = list_experiments(statuses=["completed"])
        experiment = prompt_choices(
            "Experiment",
            [(f"{x.name} #{x.number}", x) for x in reversed(experiments)],
        )
        ctx.obj["experiment_number"] = experiment.number
        return experiment.id


def experiment_id_callback(
    ctx: click.Context,
    param: click.Parameter,
    value: int,
):
    if value and "experiment_number" not in ctx.obj:
        experiment = read_experiment_by_id(value)
        ctx.obj["experiment_number"] = experiment.number
    return value


def paths_prompter(
    ctx: click.Context,
    param: click.Parameter,
    value: List[str],
) -> List[str]:
    model_source = ctx.obj.get("model_source")
    if model_source is None:
        raise click.BadOptionUsage(
            option_name="--source",
            message="Model source (`--source`) must be specified before experiment_id (`--experiment-id`).",
        )
    paths = ["/"]
    if model_source == MODEL_SOURCE_EXPERIMENT:
        experiment_number = ctx.obj.get("experiment_number")
        if experiment_number is None:
            raise click.BadOptionUsage(
                option_name="--experiment-id",
                message="Experiment id (`--experiment-id`) must be specified before paths (`--paths`).",
            )
        files = list_experiment_output_files(
            experiment_number=experiment_number,
            need_download_url=False,
            recursive=True,
            worker_number=0,
        )
        if len(files) > 0:
            paths = prompt_checkbox(
                "Paths (Press -> to select and <- to unselect)",
                choices=[(f"{x.path} {format_size(x.size)}", x.path) for x in files],
            )
            if len(paths) == 0:
                paths = ["/"]

    return paths


@model_cli.vessl_command()
@vessl_argument(
    "repository_name",
    type=click.STRING,
    required=True,
    prompter=model_repository_name_prompter,
)
@vessl_option("--model-name", type=click.STRING, help="Model name.")
@vessl_option(
    "--source",
    type=click.STRING,
    expose_value=False,
    prompter=model_source_prompter,
    callback=model_source_callback,
    help=f"{MODEL_SOURCE_EXPERIMENT} or {MODEL_SOURCE_LOCAL}.",
)
@vessl_option(
    "--experiment-id",
    type=click.INT,
    prompter=experiment_id_prompter,
    callback=experiment_id_callback,
    help=f"Experiment id to create model (only works for {MODEL_SOURCE_EXPERIMENT}).",
)
@vessl_option(
    "--path",
    type=click.STRING,
    multiple=True,
    prompter=paths_prompter,
    help=f"Path to create model (only works for {MODEL_SOURCE_EXPERIMENT}). Default: `/`",
)
@organization_name_option
@project_name_option
def create(
    repository_name: str,
    model_name: str = None,
    experiment_id: int = None,
    repository_description: str = None,
    path: str = None,
):
    model = create_model(
        repository_name=repository_name,
        repository_description=repository_description,
        experiment_id=experiment_id,
        model_name=model_name,
        paths=path,
    )
    print_success(
        f"Created '{model.model_repository.name}-{model.number}'.\n"
        f"For more info: {format_url(Endpoint.model.format(model.model_repository.organization.name, model.model_repository.name, model.number))}"
    )


@model_cli.vessl_command()
@vessl_argument(
    "repository_name",
    type=click.STRING,
    required=True,
    prompter=model_repository_name_prompter,
    callback=model_repository_name_callback,
)
@vessl_argument(
    "model_number",
    type=click.INT,
    required=True,
    prompter=model_number_prompter,
)
@vessl_argument(
    "source",
    type=click.Path(exists=True),
    required=True,
    prompter=generic_prompter("Source path"),
)
@vessl_argument(
    "dest",
    type=click.Path(),
    required=True,
    prompter=generic_prompter("Destination path", default="/"),
)
@organization_name_option
def upload(repository_name: str, model_number: int, source: str, dest: str):
    upload_model_volume_file(
        repository_name=repository_name,
        model_number=model_number,
        source_path=source,
        dest_path=dest,
    )
    print_success(f"Uploaded {source} to {dest}.")


@model_cli.vessl_command()
@vessl_argument(
    "repository_name",
    type=click.STRING,
    required=True,
    prompter=model_repository_name_prompter,
    callback=model_repository_name_callback,
)
@vessl_argument(
    "model_number",
    type=click.INT,
    required=True,
    prompter=model_number_prompter,
)
@vessl_argument(
    "source",
    type=click.Path(),
    required=True,
    prompter=generic_prompter("Source path", default="/"),
)
@vessl_argument(
    "dest",
    type=click.Path(),
    required=True,
    prompter=download_dest_path_prompter,
)
@organization_name_option
def download(repository_name: str, model_number: int, source: str, dest: str):
    download_model_volume_file(
        repository_name=repository_name,
        model_number=model_number,
        source_path=source,
        dest_path=dest,
    )
    print_success(f"Downloaded {source} to {dest}.")


@model_cli.vessl_command()
@vessl_argument(
    "repository_name",
    type=click.STRING,
    required=True,
    prompter=model_repository_name_prompter,
    callback=model_repository_name_callback,
)
@vessl_argument(
    "model_number",
    type=click.INT,
    required=True,
    prompter=model_number_prompter,
)
@vessl_argument("path", type=click.Path(), required=True, prompter=generic_prompter("File path"))
@organization_name_option
def delete_file(repository_name: str, model_number: int, path: str):
    delete_model_volume_file(
        repository_name=repository_name,
        model_number=model_number,
        path=path,
    )
    print_success(f"Deleted {path}.")


@model_cli.vessl_command()
@vessl_option(
    "--type",
    type=click.Choice(["vessl", "bento", "hf-transformers", "hf-diffusers"], case_sensitive=False),
    required=True,
    default="vessl",
    help="Type of model to register. (vessl, bento, hf-transformers, hf-diffusers)",
    prompt="Type of model to register",
)
@vessl_conditional_option(
    "--weight_name_or_path",
    type=click.STRING,
    help="Name or path of model weight. Required when the model type is huggingface",
    prompt="Name or path of model weight",
    condition=("type", ["hf-transformers", "hf-diffusers"]),
)
@vessl_option(
    "--repository_name",
    type=click.STRING,
    help="Model repository name.",
)
@vessl_option(
    "--model_number",
    type=click.INT,
    default=None,
    help="Model number (optional). If not provided, will be auto-incremented.",
)
@vessl_option(
    "--entrypoint",
    type=click.STRING,
    default=None,
    help="Entrypoint for the model. e.g. python service.py, bentoml serve model:ModelClass \n If not provided, auto-generation will be attempted.",
)
@vessl_option(
    "--python_version",
    type=click.STRING,
    default=None,
    help="Python version for the model running.",
)
@vessl_option(
    "--requirements_path",
    type=click.STRING,
    default=None,
    help="Path to requirements file. Needed for installing dependencies.",
)
@vessl_option(
    "--pyproject_path",
    type=click.STRING,
    default=None,
    help="Path to pyproject.toml file. Needed for installing dependencies.",
)
@vessl_option(
    "--cuda_version",
    type=click.STRING,
    default=None,
    help="CUDA version for the model running.",
)
@vessl_option(
    "--framework_type",
    type=click.Choice(["torch", "tensorflow"], case_sensitive=False),
    default=None,
    help="Framework type for the model running.",
)
@vessl_option(
    "--pytorch_version",
    type=click.STRING,
    default=None,
    help="PyTorch version for the model running.",
)
@vessl_option(
    "--tensorflow_version",
    type=click.STRING,
    default=None,
    help="TensorFlow version for the model running.",
)
@vessl_option(
    "-y",
    "--autocreate",
    is_flag=True,
    type=click.BOOL,
    default=False,
    help="Auto apply arguments if not provided.",
)
@vessl_option(
    "-d",
    "--dry-run",
    is_flag=True,
    type=click.BOOL,
    default=False,
    help="Dry run without registering. Only creates lockfile.",
)
@organization_name_option
def register(
    type,
    weight_name_or_path,
    repository_name,
    entrypoint,
    model_number,
    python_version,
    requirements_path,
    pyproject_path,
    cuda_version,
    framework_type,
    pytorch_version,
    tensorflow_version,
    autocreate,
    dry_run,
    **kwargs,
):
    if repository_name is None:
        repositories = list_model_repositories(limit=20)
        if len(repositories) == 0:
            if prompt_confirm("Model repository does not exist. Create?", default=True):
                repository_name = prompt_text("Model repository name")
                create_model_repository(repository_name)
                print_info(f"Created model repository {repository_name}.")
            else:
                return
        else:
            repository_name = prompt_choices("Model repository", [x.name for x in repositories])
    if os.path.exists(".vessl.model.lock") and not autocreate:
        if not prompt_confirm("Lockfile exists. Overwrite?", default=True):
            print_info("Exiting.")
            return

    if entrypoint is None:
        entrypoint = _generate_entrypoint(type, autocreate)

    if type in ("hf-transformers", "hf-diffusers"):
        # file_path = entrypoint.split(" ")[-1]
        file_path = "service.py"
        print_info(
            f"Model type is huggingface model. {file_path} file will be automatically generated."
        )
        if (
            not os.path.exists(file_path)
            or autocreate
            or prompt_confirm(f"{file_path} already exists. Overwrite?")
        ):
            _make_endpoint_file_for_hf(type, weight_name_or_path, file_path)
        else:
            print_info(f"Skipping generation of {file_path}.")

    if python_version is None:
        detected_version = ".".join(
            os.popen("python --version")
            .read()
            .strip()
            .split(" ")[1]
            .split(".")[0:2]  ## only MAJOR.MINOR version
        )
        if not autocreate:
            python_version = prompt_text(
                f"Python version? (detected: {detected_version})", default=detected_version
            )
        else:
            python_version = detected_version

    if framework_type is None:
        if autocreate:
            print_error("Framework type is required.")
            raise ValueError("Framework type is required.")
        framework_type = prompt_choices("Framework type", ["torch", "tensorflow"])

    matching_cuda_versions = []
    if framework_type == "torch" and  pytorch_version is None:
        framework_cuda_version_map = get_recent_framework_and_matchin_cuda_versions("torch")
        pytorch_version = prompt_choices("PyTorch version? (contact support@vessl.ai if you cannot find expected.)", framework_cuda_version_map.keys())
        matching_cuda_versions = framework_cuda_version_map[pytorch_version]
    elif framework_type == "tensorflow" and tensorflow_version is None:
        framework_cuda_version_map = get_recent_framework_and_matchin_cuda_versions("tensorflow")
        tensorflow_version = prompt_choices("TensorFlow version? (contact support@vessl.ai if you cannot find expected.)", framework_cuda_version_map.keys())
        matching_cuda_versions = framework_cuda_version_map[tensorflow_version]

    if cuda_version is None:
        if autocreate:
            print_error("CUDA version is required.")
            raise ValueError("CUDA version is required.")
        cuda_version = prompt_choices("CUDA version?", matching_cuda_versions)
        if cuda_version == "":
            print_error("CUDA version is required.")
            return

    (requirements_path, pyproject_path) = _get_requirement_or_pyproject(
        requirements_path, pyproject_path
    )
    if requirements_path is None and pyproject_path is None:
        print_error("Either requirements file or pyproject.toml is required.")
        return

    if model_number is None and not autocreate:
        model_number = prompt_text("Model number (optional)", default=None)

    if model_number is not None:
        try:
            existing_model = vessl_api.model_read_api(
                organization_name=_get_organization_name(**kwargs),
                repository_name=repository_name,
                number=model_number,
            )
            model_number = existing_model.number
        except Exception as e:
            print_error(f"Model {model_number} does not exist.")
            return

    if model_number is None:
        model_number = "supplied_when_registering"

    vessl_model = VesslModel.create(
        type=type,
        repository_name=repository_name,
        model_number=model_number,
        entrypoint=entrypoint,
        python_version=python_version,
        weight_name_or_path=weight_name_or_path,
        requirements_path=requirements_path,
        pyproject_path=pyproject_path,
        cuda_version=cuda_version,
        framework_type=framework_type,
        pytorch_version=pytorch_version,
        tensorflow_version=tensorflow_version,
        autocreate_repository=autocreate,
    )

    VesslModel.create_ignore_file()
    print_info("Generated ignore file.")
    if dry_run or not prompt_confirm("Register and upload current directory as model?", default=True):
        vessl_model.save_lockfile()
        print_info("Saved lockfile. Exiting.")
        return

    if model_number == "supplied_when_registering":
        print_info("Creating a new model.")
        created_model = vessl_api.model_create_api(
            organization_name=_get_organization_name(**kwargs),
            repository_name=repository_name,
        )
        model_number = created_model.number
        print_info(f"Created a new model with number {model_number}.")

    vessl_model.model_number = model_number
    vessl_model.save_lockfile()
    print_info("Lockfile saved.")

    try:
        repo = read_model_repository(repository_name=repository_name)
    except Exception as e:
        repo = None
    if repo is None:
        if autocreate:
            create_model_repository(repository_name)
        else:
            if prompt_confirm("Model repository does not exist. Create?", default=True):
                vessl_api.model_repository_create_api(
                    organization_name=_get_organization_name(**kwargs),
                    model_repository_create_api_input=ModelRepositoryCreateAPIInput(
                        name=repository_name,
                    ),
                )
            else:
                print_info("Model repository not created. Exiting.")
                print_info(
                    "Use `vessl model-repository create` to create a model repository first."
                )
                return

    src_path = f"{os.getcwd()}"
    print_info(f"{repository_name}-{model_number} {src_path} /")
    upload_model_volume_file(
        repository_name=repository_name,
        model_number=model_number,
        source_path=src_path,
        dest_path="/",
    )
    print_success(f"Registered {repository_name}-{model_number}.")


def _generate_entrypoint(input_type: str, auto) -> str:
    candidate = ""
    if input_type == "vessl":
        candidate = "vessl model launch service.py:Service -p 3000"
    elif input_type == "bento":
        candidate = "bentoml serve service:Service"
    elif input_type == "hf-transformers":
        candidate = "vessl model launch service.py:HfTransformerRunner -p 3000"
    elif input_type == "hf-diffusers":
        candidate = "vessl model launch service.py:HfDiffuserRunner -p 3000"
    if auto:
        return candidate
    if not auto and prompt_confirm(
        f"Generating entrypoint as `{candidate}`. Proceed? (No to input manually)", default=True
    ):
        target_file = "service.py" ## As every generated case uses service.py
        if not os.path.exists(target_file) and (input_type == "vessl" or input_type == "bento"):
            print_error(f"{target_file} does not exist. Please create the file first.")
            raise click.Abort()

        return candidate
    return prompt_text("Input entrypoint command")


def _get_requirement_or_pyproject(requirements_path, pyproject_path):
    if requirements_path is not None and requirements_path != "":
        return (requirements_path, None)
    if pyproject_path is not None and pyproject_path != "":
        return (None, pyproject_path)

    if requirements_path is None:
        detected_req = None
        if os.path.exists("requirements.txt"):
            detected_req = "requirements.txt"
        requirements_path = prompt_text(
            f"Path to requirements file? [detected: {detected_req}] (optional, press Enter to skip)",
            default=detected_req,
        )
        if requirements_path != "":
            return (requirements_path, None)

    if pyproject_path is None:
        detected_pyproject = None
        if os.path.exists("pyproject.toml"):
            detected_pyproject = "pyproject.toml"
        pyproject_path = prompt_text(
            f"Path to pyproject.toml file? [detected: {detected_pyproject}] (optional, press Enter to skip)",
            default=detected_pyproject,
        )
        if pyproject_path != "":
            return (None, pyproject_path)

    return (None, None)


@model_cli.vessl_command(
    help="""
    Launch model by local entrypoint. This uses .vessl.model.lock file to install dependencies.

    So you need to run `vessl model register` before launching the model.

    Directly launch model without referencing a model repository, which is the main difference from `vessl model serve` where it looks up model repository and pulls the specified model number to start.

    Also, this command only supports Runnerbase models.

    Example:

    $ vessl model launch service.py:Service
    $ vessl model launch service:Service

    """
)
@vessl_argument(
    "entrypoint",
    type=click.STRING,
    required=True,
    prompter=generic_prompter("entrypoint"),
)
@vessl_option(
    "-d",
    "--local",
    is_flag=True,
    help="Launch model locally.",
    default=False,
)
@vessl_option(
    "-p",
    "--port",
    type=click.INT,
    help="Port number to launch model. [default: 8000]",
    default=8000,
)
def launch(entrypoint: str, local: bool, port: int):
    if not os.path.exists(".vessl.model.lock"):
        print_error("Lockfile does not exist.")
        return

    filename, classname = entrypoint.split(":")
    if not os.path.exists(f"{filename}") and not os.path.exists(f"{filename}.py"):
        print_error(f"{filename} does not exist.")
        return

    lockfile = VesslModel.from_lockfile(".vessl.model.lock")
    lockfile.install_deps()
    lockfile.launch_model(filename, classname, remote=not local, port=port)



# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from vessl.openapi_client.configuration import Configuration


class ProtoLLMWorkflow(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'created_by': 'ProtoUser',
        'created_by_id': 'int',
        'created_dt': 'float',
        'default_resource_preset_id': 'int',
        'description': 'str',
        'id': 'int',
        'model_service_id': 'int',
        'name': 'str',
        'permission_config': 'V1LLMPermissionConfig',
        'status': 'str',
        'updated_dt': 'float'
    }

    attribute_map = {
        'created_by': 'created_by',
        'created_by_id': 'created_by_id',
        'created_dt': 'created_dt',
        'default_resource_preset_id': 'default_resource_preset_id',
        'description': 'description',
        'id': 'id',
        'model_service_id': 'model_service_id',
        'name': 'name',
        'permission_config': 'permission_config',
        'status': 'status',
        'updated_dt': 'updated_dt'
    }

    def __init__(self, created_by=None, created_by_id=None, created_dt=None, default_resource_preset_id=None, description=None, id=None, model_service_id=None, name=None, permission_config=None, status=None, updated_dt=None, local_vars_configuration=None):  # noqa: E501
        """ProtoLLMWorkflow - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._created_by = None
        self._created_by_id = None
        self._created_dt = None
        self._default_resource_preset_id = None
        self._description = None
        self._id = None
        self._model_service_id = None
        self._name = None
        self._permission_config = None
        self._status = None
        self._updated_dt = None
        self.discriminator = None

        if created_by is not None:
            self.created_by = created_by
        self.created_by_id = created_by_id
        self.created_dt = created_dt
        self.default_resource_preset_id = default_resource_preset_id
        if description is not None:
            self.description = description
        self.id = id
        self.model_service_id = model_service_id
        self.name = name
        if permission_config is not None:
            self.permission_config = permission_config
        self.status = status
        self.updated_dt = updated_dt

    @property
    def created_by(self):
        """Gets the created_by of this ProtoLLMWorkflow.  # noqa: E501


        :return: The created_by of this ProtoLLMWorkflow.  # noqa: E501
        :rtype: ProtoUser
        """
        return self._created_by

    @created_by.setter
    def created_by(self, created_by):
        """Sets the created_by of this ProtoLLMWorkflow.


        :param created_by: The created_by of this ProtoLLMWorkflow.  # noqa: E501
        :type created_by: ProtoUser
        """

        self._created_by = created_by

    @property
    def created_by_id(self):
        """Gets the created_by_id of this ProtoLLMWorkflow.  # noqa: E501


        :return: The created_by_id of this ProtoLLMWorkflow.  # noqa: E501
        :rtype: int
        """
        return self._created_by_id

    @created_by_id.setter
    def created_by_id(self, created_by_id):
        """Sets the created_by_id of this ProtoLLMWorkflow.


        :param created_by_id: The created_by_id of this ProtoLLMWorkflow.  # noqa: E501
        :type created_by_id: int
        """
        if self.local_vars_configuration.client_side_validation and created_by_id is None:  # noqa: E501
            raise ValueError("Invalid value for `created_by_id`, must not be `None`")  # noqa: E501

        self._created_by_id = created_by_id

    @property
    def created_dt(self):
        """Gets the created_dt of this ProtoLLMWorkflow.  # noqa: E501


        :return: The created_dt of this ProtoLLMWorkflow.  # noqa: E501
        :rtype: float
        """
        return self._created_dt

    @created_dt.setter
    def created_dt(self, created_dt):
        """Sets the created_dt of this ProtoLLMWorkflow.


        :param created_dt: The created_dt of this ProtoLLMWorkflow.  # noqa: E501
        :type created_dt: float
        """
        if self.local_vars_configuration.client_side_validation and created_dt is None:  # noqa: E501
            raise ValueError("Invalid value for `created_dt`, must not be `None`")  # noqa: E501

        self._created_dt = created_dt

    @property
    def default_resource_preset_id(self):
        """Gets the default_resource_preset_id of this ProtoLLMWorkflow.  # noqa: E501


        :return: The default_resource_preset_id of this ProtoLLMWorkflow.  # noqa: E501
        :rtype: int
        """
        return self._default_resource_preset_id

    @default_resource_preset_id.setter
    def default_resource_preset_id(self, default_resource_preset_id):
        """Sets the default_resource_preset_id of this ProtoLLMWorkflow.


        :param default_resource_preset_id: The default_resource_preset_id of this ProtoLLMWorkflow.  # noqa: E501
        :type default_resource_preset_id: int
        """
        if self.local_vars_configuration.client_side_validation and default_resource_preset_id is None:  # noqa: E501
            raise ValueError("Invalid value for `default_resource_preset_id`, must not be `None`")  # noqa: E501

        self._default_resource_preset_id = default_resource_preset_id

    @property
    def description(self):
        """Gets the description of this ProtoLLMWorkflow.  # noqa: E501


        :return: The description of this ProtoLLMWorkflow.  # noqa: E501
        :rtype: str
        """
        return self._description

    @description.setter
    def description(self, description):
        """Sets the description of this ProtoLLMWorkflow.


        :param description: The description of this ProtoLLMWorkflow.  # noqa: E501
        :type description: str
        """

        self._description = description

    @property
    def id(self):
        """Gets the id of this ProtoLLMWorkflow.  # noqa: E501


        :return: The id of this ProtoLLMWorkflow.  # noqa: E501
        :rtype: int
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this ProtoLLMWorkflow.


        :param id: The id of this ProtoLLMWorkflow.  # noqa: E501
        :type id: int
        """
        if self.local_vars_configuration.client_side_validation and id is None:  # noqa: E501
            raise ValueError("Invalid value for `id`, must not be `None`")  # noqa: E501

        self._id = id

    @property
    def model_service_id(self):
        """Gets the model_service_id of this ProtoLLMWorkflow.  # noqa: E501


        :return: The model_service_id of this ProtoLLMWorkflow.  # noqa: E501
        :rtype: int
        """
        return self._model_service_id

    @model_service_id.setter
    def model_service_id(self, model_service_id):
        """Sets the model_service_id of this ProtoLLMWorkflow.


        :param model_service_id: The model_service_id of this ProtoLLMWorkflow.  # noqa: E501
        :type model_service_id: int
        """
        if self.local_vars_configuration.client_side_validation and model_service_id is None:  # noqa: E501
            raise ValueError("Invalid value for `model_service_id`, must not be `None`")  # noqa: E501

        self._model_service_id = model_service_id

    @property
    def name(self):
        """Gets the name of this ProtoLLMWorkflow.  # noqa: E501


        :return: The name of this ProtoLLMWorkflow.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this ProtoLLMWorkflow.


        :param name: The name of this ProtoLLMWorkflow.  # noqa: E501
        :type name: str
        """
        if self.local_vars_configuration.client_side_validation and name is None:  # noqa: E501
            raise ValueError("Invalid value for `name`, must not be `None`")  # noqa: E501

        self._name = name

    @property
    def permission_config(self):
        """Gets the permission_config of this ProtoLLMWorkflow.  # noqa: E501


        :return: The permission_config of this ProtoLLMWorkflow.  # noqa: E501
        :rtype: V1LLMPermissionConfig
        """
        return self._permission_config

    @permission_config.setter
    def permission_config(self, permission_config):
        """Sets the permission_config of this ProtoLLMWorkflow.


        :param permission_config: The permission_config of this ProtoLLMWorkflow.  # noqa: E501
        :type permission_config: V1LLMPermissionConfig
        """

        self._permission_config = permission_config

    @property
    def status(self):
        """Gets the status of this ProtoLLMWorkflow.  # noqa: E501


        :return: The status of this ProtoLLMWorkflow.  # noqa: E501
        :rtype: str
        """
        return self._status

    @status.setter
    def status(self, status):
        """Sets the status of this ProtoLLMWorkflow.


        :param status: The status of this ProtoLLMWorkflow.  # noqa: E501
        :type status: str
        """
        if self.local_vars_configuration.client_side_validation and status is None:  # noqa: E501
            raise ValueError("Invalid value for `status`, must not be `None`")  # noqa: E501
        allowed_values = ["idle", "updating", "live"]  # noqa: E501
        if self.local_vars_configuration.client_side_validation and status not in allowed_values:  # noqa: E501
            raise ValueError(
                "Invalid value for `status` ({0}), must be one of {1}"  # noqa: E501
                .format(status, allowed_values)
            )

        self._status = status

    @property
    def updated_dt(self):
        """Gets the updated_dt of this ProtoLLMWorkflow.  # noqa: E501


        :return: The updated_dt of this ProtoLLMWorkflow.  # noqa: E501
        :rtype: float
        """
        return self._updated_dt

    @updated_dt.setter
    def updated_dt(self, updated_dt):
        """Sets the updated_dt of this ProtoLLMWorkflow.


        :param updated_dt: The updated_dt of this ProtoLLMWorkflow.  # noqa: E501
        :type updated_dt: float
        """
        if self.local_vars_configuration.client_side_validation and updated_dt is None:  # noqa: E501
            raise ValueError("Invalid value for `updated_dt`, must not be `None`")  # noqa: E501

        self._updated_dt = updated_dt

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ProtoLLMWorkflow):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, ProtoLLMWorkflow):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from vessl.openapi_client.configuration import Configuration


class LlmLLMWorkflowReadResponse(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'active_rollout': 'ProtoModelServiceRollout',
        'cluster': 'ProtoCluster',
        'default_resource_preset': 'ProtoResourcePreset',
        'existing_tags': 'list[ProtoTag]',
        'last_deployed_revision': 'ProtoLLMWorkflowRevision',
        'live_gateway': 'ProtoModelServiceGateway',
        'model_service': 'ProtoModelService',
        'workflow': 'ProtoLLMWorkflow'
    }

    attribute_map = {
        'active_rollout': 'ActiveRollout',
        'cluster': 'Cluster',
        'default_resource_preset': 'DefaultResourcePreset',
        'existing_tags': 'ExistingTags',
        'last_deployed_revision': 'LastDeployedRevision',
        'live_gateway': 'LiveGateway',
        'model_service': 'ModelService',
        'workflow': 'Workflow'
    }

    def __init__(self, active_rollout=None, cluster=None, default_resource_preset=None, existing_tags=None, last_deployed_revision=None, live_gateway=None, model_service=None, workflow=None, local_vars_configuration=None):  # noqa: E501
        """LlmLLMWorkflowReadResponse - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._active_rollout = None
        self._cluster = None
        self._default_resource_preset = None
        self._existing_tags = None
        self._last_deployed_revision = None
        self._live_gateway = None
        self._model_service = None
        self._workflow = None
        self.discriminator = None

        if active_rollout is not None:
            self.active_rollout = active_rollout
        if cluster is not None:
            self.cluster = cluster
        if default_resource_preset is not None:
            self.default_resource_preset = default_resource_preset
        if existing_tags is not None:
            self.existing_tags = existing_tags
        if last_deployed_revision is not None:
            self.last_deployed_revision = last_deployed_revision
        if live_gateway is not None:
            self.live_gateway = live_gateway
        if model_service is not None:
            self.model_service = model_service
        if workflow is not None:
            self.workflow = workflow

    @property
    def active_rollout(self):
        """Gets the active_rollout of this LlmLLMWorkflowReadResponse.  # noqa: E501


        :return: The active_rollout of this LlmLLMWorkflowReadResponse.  # noqa: E501
        :rtype: ProtoModelServiceRollout
        """
        return self._active_rollout

    @active_rollout.setter
    def active_rollout(self, active_rollout):
        """Sets the active_rollout of this LlmLLMWorkflowReadResponse.


        :param active_rollout: The active_rollout of this LlmLLMWorkflowReadResponse.  # noqa: E501
        :type active_rollout: ProtoModelServiceRollout
        """

        self._active_rollout = active_rollout

    @property
    def cluster(self):
        """Gets the cluster of this LlmLLMWorkflowReadResponse.  # noqa: E501


        :return: The cluster of this LlmLLMWorkflowReadResponse.  # noqa: E501
        :rtype: ProtoCluster
        """
        return self._cluster

    @cluster.setter
    def cluster(self, cluster):
        """Sets the cluster of this LlmLLMWorkflowReadResponse.


        :param cluster: The cluster of this LlmLLMWorkflowReadResponse.  # noqa: E501
        :type cluster: ProtoCluster
        """

        self._cluster = cluster

    @property
    def default_resource_preset(self):
        """Gets the default_resource_preset of this LlmLLMWorkflowReadResponse.  # noqa: E501


        :return: The default_resource_preset of this LlmLLMWorkflowReadResponse.  # noqa: E501
        :rtype: ProtoResourcePreset
        """
        return self._default_resource_preset

    @default_resource_preset.setter
    def default_resource_preset(self, default_resource_preset):
        """Sets the default_resource_preset of this LlmLLMWorkflowReadResponse.


        :param default_resource_preset: The default_resource_preset of this LlmLLMWorkflowReadResponse.  # noqa: E501
        :type default_resource_preset: ProtoResourcePreset
        """

        self._default_resource_preset = default_resource_preset

    @property
    def existing_tags(self):
        """Gets the existing_tags of this LlmLLMWorkflowReadResponse.  # noqa: E501


        :return: The existing_tags of this LlmLLMWorkflowReadResponse.  # noqa: E501
        :rtype: list[ProtoTag]
        """
        return self._existing_tags

    @existing_tags.setter
    def existing_tags(self, existing_tags):
        """Sets the existing_tags of this LlmLLMWorkflowReadResponse.


        :param existing_tags: The existing_tags of this LlmLLMWorkflowReadResponse.  # noqa: E501
        :type existing_tags: list[ProtoTag]
        """

        self._existing_tags = existing_tags

    @property
    def last_deployed_revision(self):
        """Gets the last_deployed_revision of this LlmLLMWorkflowReadResponse.  # noqa: E501


        :return: The last_deployed_revision of this LlmLLMWorkflowReadResponse.  # noqa: E501
        :rtype: ProtoLLMWorkflowRevision
        """
        return self._last_deployed_revision

    @last_deployed_revision.setter
    def last_deployed_revision(self, last_deployed_revision):
        """Sets the last_deployed_revision of this LlmLLMWorkflowReadResponse.


        :param last_deployed_revision: The last_deployed_revision of this LlmLLMWorkflowReadResponse.  # noqa: E501
        :type last_deployed_revision: ProtoLLMWorkflowRevision
        """

        self._last_deployed_revision = last_deployed_revision

    @property
    def live_gateway(self):
        """Gets the live_gateway of this LlmLLMWorkflowReadResponse.  # noqa: E501


        :return: The live_gateway of this LlmLLMWorkflowReadResponse.  # noqa: E501
        :rtype: ProtoModelServiceGateway
        """
        return self._live_gateway

    @live_gateway.setter
    def live_gateway(self, live_gateway):
        """Sets the live_gateway of this LlmLLMWorkflowReadResponse.


        :param live_gateway: The live_gateway of this LlmLLMWorkflowReadResponse.  # noqa: E501
        :type live_gateway: ProtoModelServiceGateway
        """

        self._live_gateway = live_gateway

    @property
    def model_service(self):
        """Gets the model_service of this LlmLLMWorkflowReadResponse.  # noqa: E501


        :return: The model_service of this LlmLLMWorkflowReadResponse.  # noqa: E501
        :rtype: ProtoModelService
        """
        return self._model_service

    @model_service.setter
    def model_service(self, model_service):
        """Sets the model_service of this LlmLLMWorkflowReadResponse.


        :param model_service: The model_service of this LlmLLMWorkflowReadResponse.  # noqa: E501
        :type model_service: ProtoModelService
        """

        self._model_service = model_service

    @property
    def workflow(self):
        """Gets the workflow of this LlmLLMWorkflowReadResponse.  # noqa: E501


        :return: The workflow of this LlmLLMWorkflowReadResponse.  # noqa: E501
        :rtype: ProtoLLMWorkflow
        """
        return self._workflow

    @workflow.setter
    def workflow(self, workflow):
        """Sets the workflow of this LlmLLMWorkflowReadResponse.


        :param workflow: The workflow of this LlmLLMWorkflowReadResponse.  # noqa: E501
        :type workflow: ProtoLLMWorkflow
        """

        self._workflow = workflow

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, LlmLLMWorkflowReadResponse):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, LlmLLMWorkflowReadResponse):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from vessl.openapi_client.configuration import Configuration


class ResponsePipelineStep(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'created_dt': 'datetime',
        'description': 'str',
        'emit_variables': 'list[str]',
        'external_service_spec': 'ResponsePipelineStepExternalService',
        'fail_spec': 'object',
        'id': 'int',
        'if_spec': 'ResponsePipelineStepIf',
        'input_variables': 'list[V1VariableReference]',
        'manual_input_spec': 'ResponsePipelineStepManualInput',
        'manual_judgment_spec': 'ResponsePipelineStepManualJudgment',
        'notification_spec': 'ResponsePipelineStepNotification',
        'run_spec': 'ResponsePipelineStepRun',
        'step_key': 'str',
        'title': 'str',
        'trigger_dispatch_spec': 'ResponsePipelineStepTriggerDispatch',
        'type': 'str'
    }

    attribute_map = {
        'created_dt': 'created_dt',
        'description': 'description',
        'emit_variables': 'emit_variables',
        'external_service_spec': 'external_service_spec',
        'fail_spec': 'fail_spec',
        'id': 'id',
        'if_spec': 'if_spec',
        'input_variables': 'input_variables',
        'manual_input_spec': 'manual_input_spec',
        'manual_judgment_spec': 'manual_judgment_spec',
        'notification_spec': 'notification_spec',
        'run_spec': 'run_spec',
        'step_key': 'step_key',
        'title': 'title',
        'trigger_dispatch_spec': 'trigger_dispatch_spec',
        'type': 'type'
    }

    def __init__(self, created_dt=None, description=None, emit_variables=None, external_service_spec=None, fail_spec=None, id=None, if_spec=None, input_variables=None, manual_input_spec=None, manual_judgment_spec=None, notification_spec=None, run_spec=None, step_key=None, title=None, trigger_dispatch_spec=None, type=None, local_vars_configuration=None):  # noqa: E501
        """ResponsePipelineStep - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._created_dt = None
        self._description = None
        self._emit_variables = None
        self._external_service_spec = None
        self._fail_spec = None
        self._id = None
        self._if_spec = None
        self._input_variables = None
        self._manual_input_spec = None
        self._manual_judgment_spec = None
        self._notification_spec = None
        self._run_spec = None
        self._step_key = None
        self._title = None
        self._trigger_dispatch_spec = None
        self._type = None
        self.discriminator = None

        self.created_dt = created_dt
        if description is not None:
            self.description = description
        self.emit_variables = emit_variables
        if external_service_spec is not None:
            self.external_service_spec = external_service_spec
        if fail_spec is not None:
            self.fail_spec = fail_spec
        self.id = id
        if if_spec is not None:
            self.if_spec = if_spec
        self.input_variables = input_variables
        if manual_input_spec is not None:
            self.manual_input_spec = manual_input_spec
        if manual_judgment_spec is not None:
            self.manual_judgment_spec = manual_judgment_spec
        if notification_spec is not None:
            self.notification_spec = notification_spec
        if run_spec is not None:
            self.run_spec = run_spec
        self.step_key = step_key
        self.title = title
        if trigger_dispatch_spec is not None:
            self.trigger_dispatch_spec = trigger_dispatch_spec
        if type is not None:
            self.type = type

    @property
    def created_dt(self):
        """Gets the created_dt of this ResponsePipelineStep.  # noqa: E501


        :return: The created_dt of this ResponsePipelineStep.  # noqa: E501
        :rtype: datetime
        """
        return self._created_dt

    @created_dt.setter
    def created_dt(self, created_dt):
        """Sets the created_dt of this ResponsePipelineStep.


        :param created_dt: The created_dt of this ResponsePipelineStep.  # noqa: E501
        :type created_dt: datetime
        """
        if self.local_vars_configuration.client_side_validation and created_dt is None:  # noqa: E501
            raise ValueError("Invalid value for `created_dt`, must not be `None`")  # noqa: E501

        self._created_dt = created_dt

    @property
    def description(self):
        """Gets the description of this ResponsePipelineStep.  # noqa: E501


        :return: The description of this ResponsePipelineStep.  # noqa: E501
        :rtype: str
        """
        return self._description

    @description.setter
    def description(self, description):
        """Sets the description of this ResponsePipelineStep.


        :param description: The description of this ResponsePipelineStep.  # noqa: E501
        :type description: str
        """

        self._description = description

    @property
    def emit_variables(self):
        """Gets the emit_variables of this ResponsePipelineStep.  # noqa: E501


        :return: The emit_variables of this ResponsePipelineStep.  # noqa: E501
        :rtype: list[str]
        """
        return self._emit_variables

    @emit_variables.setter
    def emit_variables(self, emit_variables):
        """Sets the emit_variables of this ResponsePipelineStep.


        :param emit_variables: The emit_variables of this ResponsePipelineStep.  # noqa: E501
        :type emit_variables: list[str]
        """
        if self.local_vars_configuration.client_side_validation and emit_variables is None:  # noqa: E501
            raise ValueError("Invalid value for `emit_variables`, must not be `None`")  # noqa: E501

        self._emit_variables = emit_variables

    @property
    def external_service_spec(self):
        """Gets the external_service_spec of this ResponsePipelineStep.  # noqa: E501


        :return: The external_service_spec of this ResponsePipelineStep.  # noqa: E501
        :rtype: ResponsePipelineStepExternalService
        """
        return self._external_service_spec

    @external_service_spec.setter
    def external_service_spec(self, external_service_spec):
        """Sets the external_service_spec of this ResponsePipelineStep.


        :param external_service_spec: The external_service_spec of this ResponsePipelineStep.  # noqa: E501
        :type external_service_spec: ResponsePipelineStepExternalService
        """

        self._external_service_spec = external_service_spec

    @property
    def fail_spec(self):
        """Gets the fail_spec of this ResponsePipelineStep.  # noqa: E501


        :return: The fail_spec of this ResponsePipelineStep.  # noqa: E501
        :rtype: object
        """
        return self._fail_spec

    @fail_spec.setter
    def fail_spec(self, fail_spec):
        """Sets the fail_spec of this ResponsePipelineStep.


        :param fail_spec: The fail_spec of this ResponsePipelineStep.  # noqa: E501
        :type fail_spec: object
        """

        self._fail_spec = fail_spec

    @property
    def id(self):
        """Gets the id of this ResponsePipelineStep.  # noqa: E501


        :return: The id of this ResponsePipelineStep.  # noqa: E501
        :rtype: int
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this ResponsePipelineStep.


        :param id: The id of this ResponsePipelineStep.  # noqa: E501
        :type id: int
        """
        if self.local_vars_configuration.client_side_validation and id is None:  # noqa: E501
            raise ValueError("Invalid value for `id`, must not be `None`")  # noqa: E501

        self._id = id

    @property
    def if_spec(self):
        """Gets the if_spec of this ResponsePipelineStep.  # noqa: E501


        :return: The if_spec of this ResponsePipelineStep.  # noqa: E501
        :rtype: ResponsePipelineStepIf
        """
        return self._if_spec

    @if_spec.setter
    def if_spec(self, if_spec):
        """Sets the if_spec of this ResponsePipelineStep.


        :param if_spec: The if_spec of this ResponsePipelineStep.  # noqa: E501
        :type if_spec: ResponsePipelineStepIf
        """

        self._if_spec = if_spec

    @property
    def input_variables(self):
        """Gets the input_variables of this ResponsePipelineStep.  # noqa: E501


        :return: The input_variables of this ResponsePipelineStep.  # noqa: E501
        :rtype: list[V1VariableReference]
        """
        return self._input_variables

    @input_variables.setter
    def input_variables(self, input_variables):
        """Sets the input_variables of this ResponsePipelineStep.


        :param input_variables: The input_variables of this ResponsePipelineStep.  # noqa: E501
        :type input_variables: list[V1VariableReference]
        """
        if self.local_vars_configuration.client_side_validation and input_variables is None:  # noqa: E501
            raise ValueError("Invalid value for `input_variables`, must not be `None`")  # noqa: E501

        self._input_variables = input_variables

    @property
    def manual_input_spec(self):
        """Gets the manual_input_spec of this ResponsePipelineStep.  # noqa: E501


        :return: The manual_input_spec of this ResponsePipelineStep.  # noqa: E501
        :rtype: ResponsePipelineStepManualInput
        """
        return self._manual_input_spec

    @manual_input_spec.setter
    def manual_input_spec(self, manual_input_spec):
        """Sets the manual_input_spec of this ResponsePipelineStep.


        :param manual_input_spec: The manual_input_spec of this ResponsePipelineStep.  # noqa: E501
        :type manual_input_spec: ResponsePipelineStepManualInput
        """

        self._manual_input_spec = manual_input_spec

    @property
    def manual_judgment_spec(self):
        """Gets the manual_judgment_spec of this ResponsePipelineStep.  # noqa: E501


        :return: The manual_judgment_spec of this ResponsePipelineStep.  # noqa: E501
        :rtype: ResponsePipelineStepManualJudgment
        """
        return self._manual_judgment_spec

    @manual_judgment_spec.setter
    def manual_judgment_spec(self, manual_judgment_spec):
        """Sets the manual_judgment_spec of this ResponsePipelineStep.


        :param manual_judgment_spec: The manual_judgment_spec of this ResponsePipelineStep.  # noqa: E501
        :type manual_judgment_spec: ResponsePipelineStepManualJudgment
        """

        self._manual_judgment_spec = manual_judgment_spec

    @property
    def notification_spec(self):
        """Gets the notification_spec of this ResponsePipelineStep.  # noqa: E501


        :return: The notification_spec of this ResponsePipelineStep.  # noqa: E501
        :rtype: ResponsePipelineStepNotification
        """
        return self._notification_spec

    @notification_spec.setter
    def notification_spec(self, notification_spec):
        """Sets the notification_spec of this ResponsePipelineStep.


        :param notification_spec: The notification_spec of this ResponsePipelineStep.  # noqa: E501
        :type notification_spec: ResponsePipelineStepNotification
        """

        self._notification_spec = notification_spec

    @property
    def run_spec(self):
        """Gets the run_spec of this ResponsePipelineStep.  # noqa: E501


        :return: The run_spec of this ResponsePipelineStep.  # noqa: E501
        :rtype: ResponsePipelineStepRun
        """
        return self._run_spec

    @run_spec.setter
    def run_spec(self, run_spec):
        """Sets the run_spec of this ResponsePipelineStep.


        :param run_spec: The run_spec of this ResponsePipelineStep.  # noqa: E501
        :type run_spec: ResponsePipelineStepRun
        """

        self._run_spec = run_spec

    @property
    def step_key(self):
        """Gets the step_key of this ResponsePipelineStep.  # noqa: E501


        :return: The step_key of this ResponsePipelineStep.  # noqa: E501
        :rtype: str
        """
        return self._step_key

    @step_key.setter
    def step_key(self, step_key):
        """Sets the step_key of this ResponsePipelineStep.


        :param step_key: The step_key of this ResponsePipelineStep.  # noqa: E501
        :type step_key: str
        """
        if self.local_vars_configuration.client_side_validation and step_key is None:  # noqa: E501
            raise ValueError("Invalid value for `step_key`, must not be `None`")  # noqa: E501

        self._step_key = step_key

    @property
    def title(self):
        """Gets the title of this ResponsePipelineStep.  # noqa: E501


        :return: The title of this ResponsePipelineStep.  # noqa: E501
        :rtype: str
        """
        return self._title

    @title.setter
    def title(self, title):
        """Sets the title of this ResponsePipelineStep.


        :param title: The title of this ResponsePipelineStep.  # noqa: E501
        :type title: str
        """
        if self.local_vars_configuration.client_side_validation and title is None:  # noqa: E501
            raise ValueError("Invalid value for `title`, must not be `None`")  # noqa: E501

        self._title = title

    @property
    def trigger_dispatch_spec(self):
        """Gets the trigger_dispatch_spec of this ResponsePipelineStep.  # noqa: E501


        :return: The trigger_dispatch_spec of this ResponsePipelineStep.  # noqa: E501
        :rtype: ResponsePipelineStepTriggerDispatch
        """
        return self._trigger_dispatch_spec

    @trigger_dispatch_spec.setter
    def trigger_dispatch_spec(self, trigger_dispatch_spec):
        """Sets the trigger_dispatch_spec of this ResponsePipelineStep.


        :param trigger_dispatch_spec: The trigger_dispatch_spec of this ResponsePipelineStep.  # noqa: E501
        :type trigger_dispatch_spec: ResponsePipelineStepTriggerDispatch
        """

        self._trigger_dispatch_spec = trigger_dispatch_spec

    @property
    def type(self):
        """Gets the type of this ResponsePipelineStep.  # noqa: E501


        :return: The type of this ResponsePipelineStep.  # noqa: E501
        :rtype: str
        """
        return self._type

    @type.setter
    def type(self, type):
        """Sets the type of this ResponsePipelineStep.


        :param type: The type of this ResponsePipelineStep.  # noqa: E501
        :type type: str
        """
        allowed_values = ["v1/run", "v1/manual_input", "v1/manual_judgment", "v1/if", "v1/fail", "v1/notification", "v1/external_service", "v1/trigger_dispatch"]  # noqa: E501
        if self.local_vars_configuration.client_side_validation and type not in allowed_values:  # noqa: E501
            raise ValueError(
                "Invalid value for `type` ({0}), must be one of {1}"  # noqa: E501
                .format(type, allowed_values)
            )

        self._type = type

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ResponsePipelineStep):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, ResponsePipelineStep):
            return True

        return self.to_dict() != other.to_dict()



import os
from typing import Callable, List, Optional, Tuple

from .file import VolumeFile


class AbstractVolume:
    def __init__(self):
        raise NotImplementedError()

    def list(self, path: str) -> List[VolumeFile]:
        """
        Lists a directory from the volume.

        The result will contain all files, including ones in subdirectories, under this directory.

        If such path does not exist (i.e. there is no such file or directory), this function will
        return an empty list.

        Arguments:
            path (str):
                Reference to the path in this volume.
        """
        raise NotImplementedError()

    def download_file(
        self,
        file: VolumeFile,
        destination: str,
        progress_callback: Optional[Callable[[int], None]] = None,
    ):
        """
        Downloads a file from the volume to local path.

        Arguments:
            file (str):
                Reference to the file in this volume.

            destination (str):
                Local path to which the file will be downloaded.

            progress_callback ((int) -> None, optional):
                Callback. This function will be called with: (downloaded bytes)
        """
        raise NotImplementedError()

    def download_directory(
        self,
        directory: str,
        destination: str,
        progress_callback: Optional[Callable[[int, int, int, int, int], None]] = None,
    ):
        """
        Downloads a directory from the volume to local path.

        Arguments:
            directory (str):
                Path of the directory in this volume.

            destination (str):
                Local path to which the file will be downloaded.  This should be a directory.
                If such path does not exist, it will be created.

            progress_callback ((int) -> None, optional):
                Callback. This function will be called with:
                    (
                        total file count,
                        1-based index of current file,
                        size of current file,
                        downloaded bytes of current file,
                        total downloaded bytes,
                    )
        """
        raise NotImplementedError()

    def _download_directory_serially(
        self,
        directory: str,
        destination: str,
        progress_callback: Optional[Callable[[int, int, int, int, int], None]] = None,
    ):
        """
        Example implementation of download_directory() that simply calls list(), and
        calls download_file() to each.
        """
        files = self.list(directory)

        total_files = len(files)
        total_bytes_downloaded = 0

        for file_idx, file in enumerate(files, 1):
            if progress_callback is not None:

                def _callback(current_bytes_downloaded: int):
                    progress_callback(
                        total_files,
                        file_idx,
                        file.size,
                        current_bytes_downloaded,
                        total_bytes_downloaded + current_bytes_downloaded,
                    )
            else:
                _callback = None

            relative_path = os.path.relpath(file.path, directory)

            self.download_file(
                file=file,
                destination=os.path.join(destination, relative_path),
                progress_callback=_callback,
            )

            total_bytes_downloaded += file.size

    def upload_file(
        self,
        local_file: str,
        destination: str,
        progress_callback: Optional[Callable[[int], None]] = None,
    ):
        """
        Uploads a local file to the volume.

        Arguments:
            local_file (str):
                Path of the local file.

            destination (str):
                Path in volume to which upload this file. If such path already exists,
                it will be overwritten.

            progress_callback ((int) -> None, optional):
                Callback. This function will be called with: (uploaded bytes)
        """
        raise NotImplementedError()

    def upload_directory(
        self,
        local_directory: str,
        destination: str,
        progress_callback: Optional[Callable[[int, int, int, int, int], None]] = None,
    ):
        """
        Uploads a local directory to the volume.

        Example:
            The following code,

                ```python
                vessl.upload_directory(
                    local_path="/a/b",
                    destination="c/d",
                )
                ```

            will upload a local file `/a/b/x/y.txt` to the volume path `c/d/x/y.txt`.

        Arguments:
            local_directory (str):
                Path of the local directory.

            destination (str):
                Path in volume to which upload this file.

            progress_callback ((int) -> None, optional):
                Callback. This function will be called with:
                    (
                        total file count,
                        1-based index of current file,
                        size of current file,
                        uploaded bytes of current file,
                        total uploaded bytes,
                    )
        """

        # [(local_path, remote_path, size_bytes), ...]
        file_list: List[Tuple[str, str, int]] = []

        for root, dirs, files in os.walk(local_directory):
            for filename in files:
                local_path = os.path.join(root, filename)
                relative_path = os.path.relpath(local_path, local_directory)
                remote_path = os.path.join(destination, relative_path)
                file_list.append((local_path, remote_path, os.stat(local_path).st_size))

        total_files = len(file_list)
        total_bytes_uploaded = 0
        for i, file in enumerate(file_list, 1):
            local_path, remote_path, size = file

            if progress_callback is not None:

                def callback(current_bytes_uploaded: int):
                    progress_callback(
                        total_files,
                        i,
                        size,
                        current_bytes_uploaded,
                        total_bytes_uploaded + current_bytes_uploaded,
                    )
            else:
                callback = None

            self.upload_file(local_path, remote_path, callback)

            total_bytes_uploaded += size

    def delete_file(self, path: str):
        """
        Deletes a file from the volume.

        Arguments:
             path (str): Path of the file to delete.
        """
        raise NotImplementedError()

    def delete_directory(self, path: str):
        """
        Deletes a directory from the volume.

        Arguments:
             path (str): Path of the directory to delete.
        """
        raise NotImplementedError()



import json
import os
import subprocess

import packaging.requirements
import toml
from packaging.specifiers import SpecifierSet

from vessl.service import _load_model_v2, _run_model_server
from vessl.util.constant import API_HOST, VESSL_IGNORE_FILE_PATH
from vessl.util.echo import print_info


class VesslModel:
    """
    This class is used to contain a model.
    You can use this class to create a lockfile, install dependencies, and launch the RunnerBase class models.
    The lockfile contains information about the model, runtime, and dependencies.

    Args:
        type_input (str): The type of model.

        repository_name (str): The name of the repository.

        model_number (str): The number of the model.

        entrypoint (str): The entrypoint of the model.

        python_version (str): The version of Python.

        weight_name_or_path (str): The name or path of the weights.

        requirements_path (str): The path to the requirements file.

        pyproject_path (str): The path to the pyproject file.

        cuda_version (str): The version of CUDA.

        pytorch_version (str): The version of PyTorch.

        tensorflow_version (str): The version of TensorFlow.

        autocreate_repository (bool): Whether to autocreate the repository.

    """

    class Runtime:
        def __init__(
            self,
            python_version=None,
            cuda_version=None,
            framework_type=None,
            pytorch_version=None,
            tensorflow_version=None,
        ):
            self.python_version = python_version
            self.cuda_version = cuda_version
            self.framework_type = framework_type
            self.pytorch_version = pytorch_version
            self.tensorflow_version = tensorflow_version

        def build(self):
            return {
                "python_version": self.python_version,
                "cuda_version": self.cuda_version,
                "framework_type": self.framework_type,
                "pytorch_version": self.pytorch_version,
                "tensorflow_version": self.tensorflow_version,
            }

    class ModelReference:
        def __init__(
            self,
            type_input=None,
            repository_name=None,
            model_number=None,
            weights_path=None,
        ):
            self.type = type_input
            self.model_repository_name = repository_name
            self.model_number = model_number
            self.weights_path = weights_path

        def build(self):
            return {
                "type": self.type,
                "model_repository_name": self.model_repository_name,
                "model_number": self.model_number,
                "weights_path": self.weights_path,
            }

    class Dependency:
        def __init__(
            self,
            requirements_path=None,
            pyproject_path=None,
        ):
            self.requirements_path = requirements_path
            self.pyproject_path = pyproject_path

        def build(self):
            return {
                "requirements_path": self.requirements_path,
                "pyproject_path": self.pyproject_path,
            }

    def __init__(
        self,
        type_input=None,
        repository_name=None,
        model_number=None,
        entrypoint=None,
        python_version=None,
        weight_name_or_path=None,
        requirements_path=None,
        pyproject_path=None,
        cuda_version=None,
        framework_type=None,
        pytorch_version=None,
        tensorflow_version=None,
        autocreate_repository=None,
    ):
        self.type = type_input
        self.repository_name = repository_name
        self.model_number = model_number
        self.entrypoint = entrypoint

        self.weight_name_or_path = weight_name_or_path

        self.python_version = python_version
        self.cuda_version = cuda_version
        self.framework_type = framework_type
        self.pytorch_version = pytorch_version
        self.tensorflow_version = tensorflow_version
        self.autocreate_repository = autocreate_repository

        self.requirements_path = requirements_path
        self.pyproject_path = pyproject_path
        if requirements_path is not None:
            self._validate_from_requirements()
        if pyproject_path is not None:
            self._validate_from_pyproject()

    def _validate_from_requirements(self):
        try:
            with open(self.requirements_path, "r") as f:
                reqs = {}
                for req_line in f:
                    req_line: str = req_line.strip()
                    if not req_line:
                        continue
                    try:
                        req = packaging.requirements.Requirement(req_line)
                    except packaging.requirements.InvalidRequirement:
                        continue
                    reqs[req.name] = req.specifier

                if "torch" in reqs:
                    _validate_version("torch", "requirements", self.pytorch_version, reqs["torch"])
                if "tensorflow" in reqs:
                    _validate_version(
                        "tensorflow", "requirements", self.tensorflow_version, reqs["tensorflow"]
                    )
                if "python" in reqs:
                    _validate_version("python", "requirements", self.python_version, reqs["python"])
        except FileNotFoundError:
            print_info("requirements.txt not found, skipping")
            pass

    def _validate_from_pyproject(self):
        try:
            with open(self.pyproject_path, "r") as f:
                data = toml.load(f)

                # pyproject deps
                deps = data.get("dependencies")
                if deps is not None:
                    deps = {k: SpecifierSet(v) for k, v in deps.items()}
                    if "torch" in deps:
                        _validate_version("torch", "pyproject", self.pytorch_version, deps["torch"])
                    if "tensorflow" in deps:
                        _validate_version(
                            "tensorlow", "pyproject", self.tensorflow_version, deps["tensorflow"]
                        )
                    if "python" in deps:
                        _validate_version(
                            "python", "pyproject", self.python_version, deps["python"]
                        )

                # poetry deps, overrides pyproject deps
                deps = data.get("tool", {}).get("poetry", {}).get("dependencies")
                if deps is not None:
                    deps = {k: SpecifierSet(v) for k, v in deps.items()}
                    if "torch" in deps:
                        _validate_version("torch", "poetry", self.pytorch_version, deps["torch"])
                    if "tensorflow" in deps:
                        _validate_version(
                            "tensorlow", "poetry", self.tensorflow_version, deps["tensorflow"]
                        )
                    if "python" in deps:
                        _validate_version("python", "poetry", self.python_version, deps["python"])
        except FileNotFoundError:
            print_info("pyproject.toml not found, skipping")
            pass

    def build_lockfile(self):
        return json.dumps(
            {
                "_comment": "This file is auto-generated by `vessl model register` command. Do not edit manually.",
                "runtime": VesslModel.Runtime(
                    self.python_version,
                    self.cuda_version,
                    self.framework_type,
                    self.pytorch_version if hasattr(self, "pytorch_version") else None,
                    self.tensorflow_version if hasattr(self, "tensorflow_version") else None,
                ).build(),
                "model": VesslModel.ModelReference(
                    self.type, self.repository_name, self.model_number, self.weight_name_or_path
                ).build(),
                "dependency": VesslModel.Dependency(
                    self.requirements_path, self.pyproject_path
                ).build(),
                "entrypoint": self.entrypoint,
            },
            indent=2,
        )

    @classmethod
    def create(
        cls,
        type: str,
        repository_name: str,
        model_number: str,
        entrypoint: str,
        python_version: str,
        weight_name_or_path: str,
        requirements_path: str,
        pyproject_path: str,
        cuda_version: str,
        framework_type: str,
        pytorch_version: str = None,
        tensorflow_version: str = None,
        autocreate_repository: bool = False,
    ):
        l = cls(
            type,
            repository_name,
            model_number,
            entrypoint,
            python_version,
            weight_name_or_path,
            requirements_path,
            pyproject_path,
            cuda_version,
            framework_type,
            pytorch_version,
            tensorflow_version,
            autocreate_repository,
        )
        return l

    def save_lockfile(self, dest_dir: str = None):
        dest_path = os.path.join(os.getcwd(), ".vessl.model.lock")
        if dest_dir != None:
            dest_path = os.path.join(dest_dir, ".vessl.model.lock")

        with open(dest_path, "w") as f:
            f.write(self.build_lockfile())

    @staticmethod
    def create_ignore_file(dest_dir: str = None):
        dest_path = VESSL_IGNORE_FILE_PATH
        if dest_dir != None:
            dest_path = os.path.join(dest_dir, ".vesslignore")

        if not os.path.exists(dest_path):
            default_vesslignore_list = [
                ".git",
                ".pytest_cache",
                ".venv",
                "__pycache__",
            ]
            with open(dest_path, "w") as f:
                f.write("\n".join(default_vesslignore_list))

    @classmethod
    def from_lockfile(cls, path: str):
        if not os.path.exists(path):
            return None
        with open(path, "r") as f:
            data = json.load(f)
            return cls(
                type_input=data["model"]["type"],
                repository_name=data["model"]["model_repository_name"],
                model_number=data["model"]["model_number"],
                entrypoint=data["entrypoint"],
                python_version=data["runtime"]["python_version"],
                requirements_path=data["dependency"]["requirements_path"],
                pyproject_path=data["dependency"]["pyproject_path"],
                cuda_version=data["runtime"]["cuda_version"],
                framework_type=data["runtime"]["framework_type"],
                pytorch_version=data["runtime"]["pytorch_version"],
                tensorflow_version=data["runtime"]["tensorflow_version"],
            )

    def install_deps(self):
        if self.requirements_path is not None:
            print_info(f"Installing dependencies from {self.requirements_path}")
            command = ["pip", "install", "-r", self.requirements_path]
            # @XXX(seokju) hardcoded for now
            if API_HOST == "https://api.dev.vssl.ai":
                command.extend(
                    [
                        "--index-url",
                        "https://test.pypi.org/simple",
                        "--extra-index-url",
                        "https://pypi.org/simple",
                        "--pre",
                    ]
                )
            subprocess.check_call(command)
        if self.pyproject_path is not None:
            print_info(f"Installing dependencies from {self.pyproject_path}")
            subprocess.check_call(["poetry", "install"])

    def launch_model(self, filename: str, classname: str, remote: bool = True, port: int = 8000):
        model_server, input_type, output_type, pass_type_param = _load_model_v2(filename, classname)

        _run_model_server(
            model_server,
            remote,
            port=port,
            api_name=f"{self.repository_name}-{self.model_number}",
            inputType=input_type if pass_type_param else None,
            outputType=output_type if pass_type_param else None,
        )


def _validate_version(library_name: str, context: str, input_version: str, version: SpecifierSet):
    if input_version is not None:
        if not version.contains(input_version):
            raise ValueError(
                f"There is a mismatch of the {library_name} version between your input({input_version}) and {context}({version})"
            )



from typing import Any

import vessl.llm.connection as connection
import vessl.llm.document as document
import vessl.llm.workflow as workflow

from vessl.openapi_client import LlmLLMWorkflowWorkloadReadConfigurationResponse, ProtoLLMWorkflowRevision, \
    V1WorkflowRevisionSpec, V1WorkflowNodeStruct, \
    ProtoLLMModelConnection, ProtoLLMVectorDBConnConfig, \
    ProtoLLMWorkloadConnectionConfiguration, V1VariableValue, LlmLLMKnowledgeIngestionConfigReadResponse, \
    LlmLLMKnowledgeRetrieverConfigReadResponse, LLMAPIKeyVerifyAPIInput, LlmLLMAPIKeyVerifyResponse, \
    LlmLLMWorkflowWorkloadReadTestConfigurationResponse, LLMWorkflowWorkloadReadTestConfigurationAPIInput, \
    ProtoRevealedSecret
from vessl.llm.knowledge import Knowledge
from vessl.llm.ingestion import IngestionConfig
from vessl.llm.retriever import RetrieverConfig
from vessl.util.exception import InvalidParamsError

from vessl.util.exception import InvalidOrganizationError
from vessl import vessl_api


def _build_workflow_from_configuration(revision_spec: V1WorkflowRevisionSpec,
                                       config: ProtoLLMWorkloadConnectionConfiguration,
                                       resolved_variables: dict[str, str]) -> workflow.Workflow:
    workflow_nodes: dict[str, workflow.WorkflowNode] = dict()
    for node_name, n in revision_spec.nodes.items():
        node: V1WorkflowNodeStruct = n
        model_connection_config: ProtoLLMModelConnection = config.model_connection_configs.get(node_name, None)
        model_connection = None
        if model_connection_config:
            model_connection = connection.ModelConnection(model_connection=model_connection_config)
        vector_db_connection: ProtoLLMVectorDBConnConfig = config.vector_db_connection_configs.get(node_name, None)
        vector_db = None
        permission = None
        if vector_db_connection:
            collection_name = config.vector_db_collections[node_name]
            vector_db = connection.VectorDBConnection(vector_db=vector_db_connection, collection_name=collection_name)
            permission = connection.PermissionConfig(config.knowledge_permission_configs[node_name])
        workflow_nodes[node_name] = workflow.WorkflowNode.from_config(node=node, model_connection=model_connection,
                                                                      vector_db=vector_db,
                                                                      permission=permission)

    workflow_edges: dict[str, list[workflow.WorkflowEdge]] = dict()
    for node_name, edges in revision_spec.edges.items():
        workflow_edges[node_name] = [workflow.WorkflowEdge.from_config(e) for e in edges]

    variables: dict[str, workflow.SingleWorkflowVariable] = dict()
    variable_specs: dict[str, V1VariableValue] = revision_spec.variables
    for variable_name, var in variable_specs.items():
        if var.source == "text":
            variables[variable_name] = workflow.SingleWorkflowVariable(
                'text', var.text, var.secret, resolved_variables.get(variable_name))

    return workflow.Workflow(nodes=workflow_nodes, edges=workflow_edges, variables=variables,
                             permission_config=config.workflow_permission_config)


def read_test_workflow_configuration(raw_spec_input: dict[str, Any]) -> workflow.Workflow:
    """
    Read the workflow configuration from api server.
    Args:
        raw_spec_input(dict): Raw spec input.
    Example:
        ```python
        vessl.read_test_workflow_configuration()
        ```
    """
    spec_input = workflow.WorkflowSpecInput.from_dict(raw_spec_input).to_v1_workflow_revision_input()
    req = LLMWorkflowWorkloadReadTestConfigurationAPIInput(spec_input=spec_input)
    workflow_test_config: LlmLLMWorkflowWorkloadReadTestConfigurationResponse = \
        vessl_api.l_lm_workflow_workload_read_test_configuration_api(
            llm_workflow_workload_read_test_configuration_api_input=req,
        )
    revision_spec: V1WorkflowRevisionSpec = workflow_test_config.workflow_spec
    connection_config: ProtoLLMWorkloadConnectionConfiguration = workflow_test_config.config
    resolved_variables: dict[str, str] = workflow_test_config.variables
    return _build_workflow_from_configuration(revision_spec, connection_config, resolved_variables)


def read_workflow_configuration() -> workflow.Workflow:
    """
    Read the workflow configuration from api server.
    Args:
        None
    Example:
        ```python
        vessl.read_workflow_configuration()
        ```
    """
    workflow_config: LlmLLMWorkflowWorkloadReadConfigurationResponse = \
        vessl_api.l_lm_workflow_workload_read_configuration_api()
    revision: ProtoLLMWorkflowRevision = workflow_config.workflow_revision
    revision_spec: V1WorkflowRevisionSpec = revision.spec
    config: ProtoLLMWorkloadConnectionConfiguration = workflow_config.config
    variables: dict[str, str] = workflow_config.variables
    return _build_workflow_from_configuration(revision_spec, config, variables)


def read_ingestion_job_configuration() -> IngestionConfig:
    config: LlmLLMKnowledgeIngestionConfigReadResponse = vessl_api.l_lm_knowledge_ingestion_config_read_api()

    knowledge = None
    if config.knowledge is not None:
        knowledge_config = config.knowledge
        knowledge = Knowledge(knowledge=knowledge_config)

    vector_db = None
    if config.vector_db_connection is not None and config.knowledge is not None:
        vector_db_config = config.vector_db_connection
        vector_db = connection.VectorDBConnection(vector_db=vector_db_config,
                                                  collection_name=config.knowledge.vectordb_collection_name)

    model_connection = None
    if config.model_connection is not None:
        model_connection_config: ProtoLLMModelConnection = config.model_connection
        model_connection = connection.ModelConnection(model_connection=model_connection_config)

    documents = []
    if config.documents is not None:
        documents = [document.Document(doc) for doc in config.documents]

    return IngestionConfig(llm_knowledge=knowledge, llm_model=model_connection, vector_db=vector_db,
                           documents=documents)


def read_retriever_configuration() -> RetrieverConfig:
    config: LlmLLMKnowledgeRetrieverConfigReadResponse = vessl_api.l_lm_knowledge_retriever_config_read_api()

    knowledge = None
    if config.knowledge is not None:
        knowledge_config = config.knowledge
        knowledge = Knowledge(knowledge=knowledge_config)

    vector_db = None
    if config.vector_db_connection is not None and config.knowledge is not None:
        vector_db_config = config.vector_db_connection
        vector_db = connection.VectorDBConnection(vector_db=vector_db_config,
                                                  collection_name=config.knowledge.vectordb_collection_name)

    model_connection = None
    if config.model_connection is not None:
        model_connection_config: ProtoLLMModelConnection = config.model_connection
        model_connection = connection.ModelConnection(model_connection=model_connection_config)

    return RetrieverConfig(llm_knowledge=knowledge, llm_model=model_connection, vector_db=vector_db)


def update_job_status(job_number: int, message: str, status: str, **kwargs):
    """
    Update the job status.
    Args:
        job_number(int): Job number.
        message(str): Message.
        status(str): Status.

    Example:
        ```python
        vessl.update_job_status(
            job_number=1,
            message="job completed",
            status="completed",
            organization_name="foo",
            knowledge_name="bar",
        )
        ```
    """
    return vessl_api.l_lm_knowledge_ingestion_job_status_update_api(
        knowledge_name=kwargs.get("knowledge_name"),
        organization_name=_get_organization_name(**kwargs),
        job_number=job_number,
        llm_knowledge_ingestion_job_status_update_api_input={
            "status": status,
            "message": message,
        }
    )


def verify_llm_api_key(api_key: str) -> connection.UserGroup:
    """
    Verify the api key.
    Args:
        api_key(str): Decrypted API key.

    Example:
        ```python
        vessl.verify_llm__api_key(api_key="foo")
        ```
    """

    resp: LlmLLMAPIKeyVerifyResponse = vessl_api.l_lmapi_key_verify_api(
        llmapi_key_verify_api_input=LLMAPIKeyVerifyAPIInput(api_key=api_key))

    if not resp.is_verified:
        raise InvalidParamsError("API key verification failed.")

    return connection.UserGroup(resp.user_group)


def _get_organization_name(**kwargs) -> str:
    organization_name = kwargs.get("organization_name")
    if organization_name is not None:
        return organization_name
    if vessl_api.organization is not None:
        return vessl_api.organization.name
    if vessl_api.run_execution_default_organization is not None:
        return vessl_api.run_execution_default_organization
    raise InvalidOrganizationError("No organization selected.")



from typing import Any, List

import click

TAB_SIZE = 2


def format_key(key: str) -> str:
    return click.style(key, fg="green")


def format_data(data: Any, depth: int) -> List[str]:
    indent = " " * (TAB_SIZE * depth + 1)

    if not isinstance(data, dict):
        # `data` is primitive type
        return [f"{indent}{data}"]

    lines = []
    for key, value in data.items():
        if isinstance(value, dict):
            lines.append(f"{indent}{format_key(key)}")
            lines.extend(format_data(value, depth + 1))
        elif isinstance(value, list):
            lines.append(f"{indent}{format_key(key)}")
            for x in value:
                # TODO: find a more elegant way. Perhaps yaml library?
                additional_lines = format_data(x, depth + 1)
                additional_lines[0] = additional_lines[0].lstrip()
                additional_lines[0] = f"{indent}- {additional_lines[0]}"
                lines.extend(additional_lines)
        else:
            lines.append(f"{indent}{format_key(key)} {value}")
    return lines


def format_size(value: int, suffix="B") -> str:
    if value == 0:
        return "0 B"

    for unit in ["", "K", "M", "G", "T", "P", "E", "Z"]:
        if abs(value) < 1024.0:
            return f"{value:.1f} {unit}{suffix}"
        value /= 1024.0
    return f"{value:.1f} Yi{suffix}"


def format_bool(value: bool, true_value: str = "Y", false_value: str = "N") -> str:
    return true_value if value else false_value


def format_string(value: str, null_value: str = "None", empty_value: str = "-") -> str:
    if value is None:
        return null_value
    if not value:
        return empty_value
    return value


def format_url(url: str) -> str:
    """Prints in url format

    Replace whitespaces with %20. More url rules could be added.
    """
    return url.replace(" ", "%20")


def sizeof_fmt(num, suffix="B"):
    for unit in ["", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"]:
        if abs(num) < 1024.0:
            return f"{num:3.1f}{unit}{suffix}"
        num /= 1024.0
    return f"{num:.1f}Yi{suffix}"



from __future__ import absolute_import

# flake8: noqa

# import apis into api package
from vessl.openapi_client.api.apiv1_api import APIV1Api
from vessl.openapi_client.api.aws_marketplace_v1_api import AWSMarketplaceV1Api
from vessl.openapi_client.api.exec_apiv1_api import ExecAPIV1Api
from vessl.openapi_client.api.ssoapiv1_api import SSOAPIV1Api



# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from vessl.openapi_client.configuration import Configuration


class PrometheusqueryLatestNodeMetricSample(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'cpu': 'PrometheusqueryNodeResourceMetricSample',
        'disk': 'PrometheusqueryNodeResourceMetricSample',
        'gpu': 'PrometheusqueryNodeResourceMetricSample',
        'gpu_slice': 'PrometheusqueryNodeResourceMetricSample',
        'memory': 'PrometheusqueryNodeResourceMetricSample'
    }

    attribute_map = {
        'cpu': 'cpu',
        'disk': 'disk',
        'gpu': 'gpu',
        'gpu_slice': 'gpu_slice',
        'memory': 'memory'
    }

    def __init__(self, cpu=None, disk=None, gpu=None, gpu_slice=None, memory=None, local_vars_configuration=None):  # noqa: E501
        """PrometheusqueryLatestNodeMetricSample - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._cpu = None
        self._disk = None
        self._gpu = None
        self._gpu_slice = None
        self._memory = None
        self.discriminator = None

        self.cpu = cpu
        self.disk = disk
        self.gpu = gpu
        self.gpu_slice = gpu_slice
        self.memory = memory

    @property
    def cpu(self):
        """Gets the cpu of this PrometheusqueryLatestNodeMetricSample.  # noqa: E501


        :return: The cpu of this PrometheusqueryLatestNodeMetricSample.  # noqa: E501
        :rtype: PrometheusqueryNodeResourceMetricSample
        """
        return self._cpu

    @cpu.setter
    def cpu(self, cpu):
        """Sets the cpu of this PrometheusqueryLatestNodeMetricSample.


        :param cpu: The cpu of this PrometheusqueryLatestNodeMetricSample.  # noqa: E501
        :type cpu: PrometheusqueryNodeResourceMetricSample
        """
        if self.local_vars_configuration.client_side_validation and cpu is None:  # noqa: E501
            raise ValueError("Invalid value for `cpu`, must not be `None`")  # noqa: E501

        self._cpu = cpu

    @property
    def disk(self):
        """Gets the disk of this PrometheusqueryLatestNodeMetricSample.  # noqa: E501


        :return: The disk of this PrometheusqueryLatestNodeMetricSample.  # noqa: E501
        :rtype: PrometheusqueryNodeResourceMetricSample
        """
        return self._disk

    @disk.setter
    def disk(self, disk):
        """Sets the disk of this PrometheusqueryLatestNodeMetricSample.


        :param disk: The disk of this PrometheusqueryLatestNodeMetricSample.  # noqa: E501
        :type disk: PrometheusqueryNodeResourceMetricSample
        """
        if self.local_vars_configuration.client_side_validation and disk is None:  # noqa: E501
            raise ValueError("Invalid value for `disk`, must not be `None`")  # noqa: E501

        self._disk = disk

    @property
    def gpu(self):
        """Gets the gpu of this PrometheusqueryLatestNodeMetricSample.  # noqa: E501


        :return: The gpu of this PrometheusqueryLatestNodeMetricSample.  # noqa: E501
        :rtype: PrometheusqueryNodeResourceMetricSample
        """
        return self._gpu

    @gpu.setter
    def gpu(self, gpu):
        """Sets the gpu of this PrometheusqueryLatestNodeMetricSample.


        :param gpu: The gpu of this PrometheusqueryLatestNodeMetricSample.  # noqa: E501
        :type gpu: PrometheusqueryNodeResourceMetricSample
        """
        if self.local_vars_configuration.client_side_validation and gpu is None:  # noqa: E501
            raise ValueError("Invalid value for `gpu`, must not be `None`")  # noqa: E501

        self._gpu = gpu

    @property
    def gpu_slice(self):
        """Gets the gpu_slice of this PrometheusqueryLatestNodeMetricSample.  # noqa: E501


        :return: The gpu_slice of this PrometheusqueryLatestNodeMetricSample.  # noqa: E501
        :rtype: PrometheusqueryNodeResourceMetricSample
        """
        return self._gpu_slice

    @gpu_slice.setter
    def gpu_slice(self, gpu_slice):
        """Sets the gpu_slice of this PrometheusqueryLatestNodeMetricSample.


        :param gpu_slice: The gpu_slice of this PrometheusqueryLatestNodeMetricSample.  # noqa: E501
        :type gpu_slice: PrometheusqueryNodeResourceMetricSample
        """
        if self.local_vars_configuration.client_side_validation and gpu_slice is None:  # noqa: E501
            raise ValueError("Invalid value for `gpu_slice`, must not be `None`")  # noqa: E501

        self._gpu_slice = gpu_slice

    @property
    def memory(self):
        """Gets the memory of this PrometheusqueryLatestNodeMetricSample.  # noqa: E501


        :return: The memory of this PrometheusqueryLatestNodeMetricSample.  # noqa: E501
        :rtype: PrometheusqueryNodeResourceMetricSample
        """
        return self._memory

    @memory.setter
    def memory(self, memory):
        """Sets the memory of this PrometheusqueryLatestNodeMetricSample.


        :param memory: The memory of this PrometheusqueryLatestNodeMetricSample.  # noqa: E501
        :type memory: PrometheusqueryNodeResourceMetricSample
        """
        if self.local_vars_configuration.client_side_validation and memory is None:  # noqa: E501
            raise ValueError("Invalid value for `memory`, must not be `None`")  # noqa: E501

        self._memory = memory

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, PrometheusqueryLatestNodeMetricSample):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, PrometheusqueryLatestNodeMetricSample):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from vessl.openapi_client.configuration import Configuration


class OrmUser(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'created_dt': 'datetime',
        'default_organization_id': 'int',
        'display_name': 'str',
        'edges': 'OrmUserEdges',
        'email': 'str',
        'encrypted_git_ssh_key': 'str',
        'extra_info': 'dict[str, object]',
        'first_time_credit_given': 'bool',
        'git_ssh_key_name': 'str',
        'id': 'int',
        'immutable_slug': 'str',
        'is_email_verified': 'bool',
        'is_locked': 'bool',
        'is_pending': 'bool',
        'is_sso': 'bool',
        'last_login': 'datetime',
        'last_password_changed_dt': 'datetime',
        'marketing_consent_dt': 'datetime',
        'notification_config': 'dict[str, bool]',
        'status': 'str',
        'updated_dt': 'datetime',
        'username': 'str'
    }

    attribute_map = {
        'created_dt': 'created_dt',
        'default_organization_id': 'default_organization_id',
        'display_name': 'display_name',
        'edges': 'edges',
        'email': 'email',
        'encrypted_git_ssh_key': 'encrypted_git_ssh_key',
        'extra_info': 'extra_info',
        'first_time_credit_given': 'first_time_credit_given',
        'git_ssh_key_name': 'git_ssh_key_name',
        'id': 'id',
        'immutable_slug': 'immutable_slug',
        'is_email_verified': 'is_email_verified',
        'is_locked': 'is_locked',
        'is_pending': 'is_pending',
        'is_sso': 'is_sso',
        'last_login': 'last_login',
        'last_password_changed_dt': 'last_password_changed_dt',
        'marketing_consent_dt': 'marketing_consent_dt',
        'notification_config': 'notification_config',
        'status': 'status',
        'updated_dt': 'updated_dt',
        'username': 'username'
    }

    def __init__(self, created_dt=None, default_organization_id=None, display_name=None, edges=None, email=None, encrypted_git_ssh_key=None, extra_info=None, first_time_credit_given=None, git_ssh_key_name=None, id=None, immutable_slug=None, is_email_verified=None, is_locked=None, is_pending=None, is_sso=None, last_login=None, last_password_changed_dt=None, marketing_consent_dt=None, notification_config=None, status=None, updated_dt=None, username=None, local_vars_configuration=None):  # noqa: E501
        """OrmUser - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._created_dt = None
        self._default_organization_id = None
        self._display_name = None
        self._edges = None
        self._email = None
        self._encrypted_git_ssh_key = None
        self._extra_info = None
        self._first_time_credit_given = None
        self._git_ssh_key_name = None
        self._id = None
        self._immutable_slug = None
        self._is_email_verified = None
        self._is_locked = None
        self._is_pending = None
        self._is_sso = None
        self._last_login = None
        self._last_password_changed_dt = None
        self._marketing_consent_dt = None
        self._notification_config = None
        self._status = None
        self._updated_dt = None
        self._username = None
        self.discriminator = None

        if created_dt is not None:
            self.created_dt = created_dt
        self.default_organization_id = default_organization_id
        if display_name is not None:
            self.display_name = display_name
        if edges is not None:
            self.edges = edges
        if email is not None:
            self.email = email
        self.encrypted_git_ssh_key = encrypted_git_ssh_key
        if extra_info is not None:
            self.extra_info = extra_info
        if first_time_credit_given is not None:
            self.first_time_credit_given = first_time_credit_given
        self.git_ssh_key_name = git_ssh_key_name
        if id is not None:
            self.id = id
        if immutable_slug is not None:
            self.immutable_slug = immutable_slug
        if is_email_verified is not None:
            self.is_email_verified = is_email_verified
        if is_locked is not None:
            self.is_locked = is_locked
        if is_pending is not None:
            self.is_pending = is_pending
        if is_sso is not None:
            self.is_sso = is_sso
        self.last_login = last_login
        self.last_password_changed_dt = last_password_changed_dt
        self.marketing_consent_dt = marketing_consent_dt
        if notification_config is not None:
            self.notification_config = notification_config
        if status is not None:
            self.status = status
        if updated_dt is not None:
            self.updated_dt = updated_dt
        if username is not None:
            self.username = username

    @property
    def created_dt(self):
        """Gets the created_dt of this OrmUser.  # noqa: E501


        :return: The created_dt of this OrmUser.  # noqa: E501
        :rtype: datetime
        """
        return self._created_dt

    @created_dt.setter
    def created_dt(self, created_dt):
        """Sets the created_dt of this OrmUser.


        :param created_dt: The created_dt of this OrmUser.  # noqa: E501
        :type created_dt: datetime
        """

        self._created_dt = created_dt

    @property
    def default_organization_id(self):
        """Gets the default_organization_id of this OrmUser.  # noqa: E501


        :return: The default_organization_id of this OrmUser.  # noqa: E501
        :rtype: int
        """
        return self._default_organization_id

    @default_organization_id.setter
    def default_organization_id(self, default_organization_id):
        """Sets the default_organization_id of this OrmUser.


        :param default_organization_id: The default_organization_id of this OrmUser.  # noqa: E501
        :type default_organization_id: int
        """

        self._default_organization_id = default_organization_id

    @property
    def display_name(self):
        """Gets the display_name of this OrmUser.  # noqa: E501


        :return: The display_name of this OrmUser.  # noqa: E501
        :rtype: str
        """
        return self._display_name

    @display_name.setter
    def display_name(self, display_name):
        """Sets the display_name of this OrmUser.


        :param display_name: The display_name of this OrmUser.  # noqa: E501
        :type display_name: str
        """

        self._display_name = display_name

    @property
    def edges(self):
        """Gets the edges of this OrmUser.  # noqa: E501


        :return: The edges of this OrmUser.  # noqa: E501
        :rtype: OrmUserEdges
        """
        return self._edges

    @edges.setter
    def edges(self, edges):
        """Sets the edges of this OrmUser.


        :param edges: The edges of this OrmUser.  # noqa: E501
        :type edges: OrmUserEdges
        """

        self._edges = edges

    @property
    def email(self):
        """Gets the email of this OrmUser.  # noqa: E501


        :return: The email of this OrmUser.  # noqa: E501
        :rtype: str
        """
        return self._email

    @email.setter
    def email(self, email):
        """Sets the email of this OrmUser.


        :param email: The email of this OrmUser.  # noqa: E501
        :type email: str
        """

        self._email = email

    @property
    def encrypted_git_ssh_key(self):
        """Gets the encrypted_git_ssh_key of this OrmUser.  # noqa: E501


        :return: The encrypted_git_ssh_key of this OrmUser.  # noqa: E501
        :rtype: str
        """
        return self._encrypted_git_ssh_key

    @encrypted_git_ssh_key.setter
    def encrypted_git_ssh_key(self, encrypted_git_ssh_key):
        """Sets the encrypted_git_ssh_key of this OrmUser.


        :param encrypted_git_ssh_key: The encrypted_git_ssh_key of this OrmUser.  # noqa: E501
        :type encrypted_git_ssh_key: str
        """

        self._encrypted_git_ssh_key = encrypted_git_ssh_key

    @property
    def extra_info(self):
        """Gets the extra_info of this OrmUser.  # noqa: E501


        :return: The extra_info of this OrmUser.  # noqa: E501
        :rtype: dict[str, object]
        """
        return self._extra_info

    @extra_info.setter
    def extra_info(self, extra_info):
        """Sets the extra_info of this OrmUser.


        :param extra_info: The extra_info of this OrmUser.  # noqa: E501
        :type extra_info: dict[str, object]
        """

        self._extra_info = extra_info

    @property
    def first_time_credit_given(self):
        """Gets the first_time_credit_given of this OrmUser.  # noqa: E501


        :return: The first_time_credit_given of this OrmUser.  # noqa: E501
        :rtype: bool
        """
        return self._first_time_credit_given

    @first_time_credit_given.setter
    def first_time_credit_given(self, first_time_credit_given):
        """Sets the first_time_credit_given of this OrmUser.


        :param first_time_credit_given: The first_time_credit_given of this OrmUser.  # noqa: E501
        :type first_time_credit_given: bool
        """

        self._first_time_credit_given = first_time_credit_given

    @property
    def git_ssh_key_name(self):
        """Gets the git_ssh_key_name of this OrmUser.  # noqa: E501


        :return: The git_ssh_key_name of this OrmUser.  # noqa: E501
        :rtype: str
        """
        return self._git_ssh_key_name

    @git_ssh_key_name.setter
    def git_ssh_key_name(self, git_ssh_key_name):
        """Sets the git_ssh_key_name of this OrmUser.


        :param git_ssh_key_name: The git_ssh_key_name of this OrmUser.  # noqa: E501
        :type git_ssh_key_name: str
        """

        self._git_ssh_key_name = git_ssh_key_name

    @property
    def id(self):
        """Gets the id of this OrmUser.  # noqa: E501


        :return: The id of this OrmUser.  # noqa: E501
        :rtype: int
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this OrmUser.


        :param id: The id of this OrmUser.  # noqa: E501
        :type id: int
        """

        self._id = id

    @property
    def immutable_slug(self):
        """Gets the immutable_slug of this OrmUser.  # noqa: E501


        :return: The immutable_slug of this OrmUser.  # noqa: E501
        :rtype: str
        """
        return self._immutable_slug

    @immutable_slug.setter
    def immutable_slug(self, immutable_slug):
        """Sets the immutable_slug of this OrmUser.


        :param immutable_slug: The immutable_slug of this OrmUser.  # noqa: E501
        :type immutable_slug: str
        """

        self._immutable_slug = immutable_slug

    @property
    def is_email_verified(self):
        """Gets the is_email_verified of this OrmUser.  # noqa: E501


        :return: The is_email_verified of this OrmUser.  # noqa: E501
        :rtype: bool
        """
        return self._is_email_verified

    @is_email_verified.setter
    def is_email_verified(self, is_email_verified):
        """Sets the is_email_verified of this OrmUser.


        :param is_email_verified: The is_email_verified of this OrmUser.  # noqa: E501
        :type is_email_verified: bool
        """

        self._is_email_verified = is_email_verified

    @property
    def is_locked(self):
        """Gets the is_locked of this OrmUser.  # noqa: E501


        :return: The is_locked of this OrmUser.  # noqa: E501
        :rtype: bool
        """
        return self._is_locked

    @is_locked.setter
    def is_locked(self, is_locked):
        """Sets the is_locked of this OrmUser.


        :param is_locked: The is_locked of this OrmUser.  # noqa: E501
        :type is_locked: bool
        """

        self._is_locked = is_locked

    @property
    def is_pending(self):
        """Gets the is_pending of this OrmUser.  # noqa: E501


        :return: The is_pending of this OrmUser.  # noqa: E501
        :rtype: bool
        """
        return self._is_pending

    @is_pending.setter
    def is_pending(self, is_pending):
        """Sets the is_pending of this OrmUser.


        :param is_pending: The is_pending of this OrmUser.  # noqa: E501
        :type is_pending: bool
        """

        self._is_pending = is_pending

    @property
    def is_sso(self):
        """Gets the is_sso of this OrmUser.  # noqa: E501


        :return: The is_sso of this OrmUser.  # noqa: E501
        :rtype: bool
        """
        return self._is_sso

    @is_sso.setter
    def is_sso(self, is_sso):
        """Sets the is_sso of this OrmUser.


        :param is_sso: The is_sso of this OrmUser.  # noqa: E501
        :type is_sso: bool
        """

        self._is_sso = is_sso

    @property
    def last_login(self):
        """Gets the last_login of this OrmUser.  # noqa: E501


        :return: The last_login of this OrmUser.  # noqa: E501
        :rtype: datetime
        """
        return self._last_login

    @last_login.setter
    def last_login(self, last_login):
        """Sets the last_login of this OrmUser.


        :param last_login: The last_login of this OrmUser.  # noqa: E501
        :type last_login: datetime
        """

        self._last_login = last_login

    @property
    def last_password_changed_dt(self):
        """Gets the last_password_changed_dt of this OrmUser.  # noqa: E501


        :return: The last_password_changed_dt of this OrmUser.  # noqa: E501
        :rtype: datetime
        """
        return self._last_password_changed_dt

    @last_password_changed_dt.setter
    def last_password_changed_dt(self, last_password_changed_dt):
        """Sets the last_password_changed_dt of this OrmUser.


        :param last_password_changed_dt: The last_password_changed_dt of this OrmUser.  # noqa: E501
        :type last_password_changed_dt: datetime
        """

        self._last_password_changed_dt = last_password_changed_dt

    @property
    def marketing_consent_dt(self):
        """Gets the marketing_consent_dt of this OrmUser.  # noqa: E501


        :return: The marketing_consent_dt of this OrmUser.  # noqa: E501
        :rtype: datetime
        """
        return self._marketing_consent_dt

    @marketing_consent_dt.setter
    def marketing_consent_dt(self, marketing_consent_dt):
        """Sets the marketing_consent_dt of this OrmUser.


        :param marketing_consent_dt: The marketing_consent_dt of this OrmUser.  # noqa: E501
        :type marketing_consent_dt: datetime
        """

        self._marketing_consent_dt = marketing_consent_dt

    @property
    def notification_config(self):
        """Gets the notification_config of this OrmUser.  # noqa: E501


        :return: The notification_config of this OrmUser.  # noqa: E501
        :rtype: dict[str, bool]
        """
        return self._notification_config

    @notification_config.setter
    def notification_config(self, notification_config):
        """Sets the notification_config of this OrmUser.


        :param notification_config: The notification_config of this OrmUser.  # noqa: E501
        :type notification_config: dict[str, bool]
        """

        self._notification_config = notification_config

    @property
    def status(self):
        """Gets the status of this OrmUser.  # noqa: E501


        :return: The status of this OrmUser.  # noqa: E501
        :rtype: str
        """
        return self._status

    @status.setter
    def status(self, status):
        """Sets the status of this OrmUser.


        :param status: The status of this OrmUser.  # noqa: E501
        :type status: str
        """

        self._status = status

    @property
    def updated_dt(self):
        """Gets the updated_dt of this OrmUser.  # noqa: E501


        :return: The updated_dt of this OrmUser.  # noqa: E501
        :rtype: datetime
        """
        return self._updated_dt

    @updated_dt.setter
    def updated_dt(self, updated_dt):
        """Sets the updated_dt of this OrmUser.


        :param updated_dt: The updated_dt of this OrmUser.  # noqa: E501
        :type updated_dt: datetime
        """

        self._updated_dt = updated_dt

    @property
    def username(self):
        """Gets the username of this OrmUser.  # noqa: E501


        :return: The username of this OrmUser.  # noqa: E501
        :rtype: str
        """
        return self._username

    @username.setter
    def username(self, username):
        """Sets the username of this OrmUser.


        :param username: The username of this OrmUser.  # noqa: E501
        :type username: str
        """

        self._username = username

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, OrmUser):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, OrmUser):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from vessl.openapi_client.configuration import Configuration


class V1ModelServiceRevisionSpec(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'autoscaling': 'V1Autoscaling',
        'command': 'list[V1Command]',
        'env_vars': 'dict[str, V1EnvVar]',
        'expose': 'int',
        'monitoring': 'list[V1PortPath]',
        'node_ids': 'list[int]',
        'openapi': 'V1OpenAPIForm',
        'ports': 'list[V1Port]',
        'pre_created_service_account_name': 'str',
        'readiness_probe': 'V1ReadinessProbe',
        'runner_type': 'str',
        'service_account': 'V1ServiceAccount',
        'volume_claim_mounts': 'dict[str, V1VolumeClaimMount]',
        'volume_v2_mounts': 'list[V1VolumeV2Mount]'
    }

    attribute_map = {
        'autoscaling': 'autoscaling',
        'command': 'command',
        'env_vars': 'env_vars',
        'expose': 'expose',
        'monitoring': 'monitoring',
        'node_ids': 'node_ids',
        'openapi': 'openapi',
        'ports': 'ports',
        'pre_created_service_account_name': 'pre_created_service_account_name',
        'readiness_probe': 'readiness_probe',
        'runner_type': 'runner_type',
        'service_account': 'service_account',
        'volume_claim_mounts': 'volume_claim_mounts',
        'volume_v2_mounts': 'volume_v2_mounts'
    }

    def __init__(self, autoscaling=None, command=None, env_vars=None, expose=None, monitoring=None, node_ids=None, openapi=None, ports=None, pre_created_service_account_name=None, readiness_probe=None, runner_type=None, service_account=None, volume_claim_mounts=None, volume_v2_mounts=None, local_vars_configuration=None):  # noqa: E501
        """V1ModelServiceRevisionSpec - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._autoscaling = None
        self._command = None
        self._env_vars = None
        self._expose = None
        self._monitoring = None
        self._node_ids = None
        self._openapi = None
        self._ports = None
        self._pre_created_service_account_name = None
        self._readiness_probe = None
        self._runner_type = None
        self._service_account = None
        self._volume_claim_mounts = None
        self._volume_v2_mounts = None
        self.discriminator = None

        if autoscaling is not None:
            self.autoscaling = autoscaling
        if command is not None:
            self.command = command
        if env_vars is not None:
            self.env_vars = env_vars
        if expose is not None:
            self.expose = expose
        if monitoring is not None:
            self.monitoring = monitoring
        if node_ids is not None:
            self.node_ids = node_ids
        if openapi is not None:
            self.openapi = openapi
        if ports is not None:
            self.ports = ports
        self.pre_created_service_account_name = pre_created_service_account_name
        if readiness_probe is not None:
            self.readiness_probe = readiness_probe
        self.runner_type = runner_type
        if service_account is not None:
            self.service_account = service_account
        if volume_claim_mounts is not None:
            self.volume_claim_mounts = volume_claim_mounts
        if volume_v2_mounts is not None:
            self.volume_v2_mounts = volume_v2_mounts

    @property
    def autoscaling(self):
        """Gets the autoscaling of this V1ModelServiceRevisionSpec.  # noqa: E501


        :return: The autoscaling of this V1ModelServiceRevisionSpec.  # noqa: E501
        :rtype: V1Autoscaling
        """
        return self._autoscaling

    @autoscaling.setter
    def autoscaling(self, autoscaling):
        """Sets the autoscaling of this V1ModelServiceRevisionSpec.


        :param autoscaling: The autoscaling of this V1ModelServiceRevisionSpec.  # noqa: E501
        :type autoscaling: V1Autoscaling
        """

        self._autoscaling = autoscaling

    @property
    def command(self):
        """Gets the command of this V1ModelServiceRevisionSpec.  # noqa: E501


        :return: The command of this V1ModelServiceRevisionSpec.  # noqa: E501
        :rtype: list[V1Command]
        """
        return self._command

    @command.setter
    def command(self, command):
        """Sets the command of this V1ModelServiceRevisionSpec.


        :param command: The command of this V1ModelServiceRevisionSpec.  # noqa: E501
        :type command: list[V1Command]
        """

        self._command = command

    @property
    def env_vars(self):
        """Gets the env_vars of this V1ModelServiceRevisionSpec.  # noqa: E501


        :return: The env_vars of this V1ModelServiceRevisionSpec.  # noqa: E501
        :rtype: dict[str, V1EnvVar]
        """
        return self._env_vars

    @env_vars.setter
    def env_vars(self, env_vars):
        """Sets the env_vars of this V1ModelServiceRevisionSpec.


        :param env_vars: The env_vars of this V1ModelServiceRevisionSpec.  # noqa: E501
        :type env_vars: dict[str, V1EnvVar]
        """

        self._env_vars = env_vars

    @property
    def expose(self):
        """Gets the expose of this V1ModelServiceRevisionSpec.  # noqa: E501


        :return: The expose of this V1ModelServiceRevisionSpec.  # noqa: E501
        :rtype: int
        """
        return self._expose

    @expose.setter
    def expose(self, expose):
        """Sets the expose of this V1ModelServiceRevisionSpec.


        :param expose: The expose of this V1ModelServiceRevisionSpec.  # noqa: E501
        :type expose: int
        """

        self._expose = expose

    @property
    def monitoring(self):
        """Gets the monitoring of this V1ModelServiceRevisionSpec.  # noqa: E501


        :return: The monitoring of this V1ModelServiceRevisionSpec.  # noqa: E501
        :rtype: list[V1PortPath]
        """
        return self._monitoring

    @monitoring.setter
    def monitoring(self, monitoring):
        """Sets the monitoring of this V1ModelServiceRevisionSpec.


        :param monitoring: The monitoring of this V1ModelServiceRevisionSpec.  # noqa: E501
        :type monitoring: list[V1PortPath]
        """

        self._monitoring = monitoring

    @property
    def node_ids(self):
        """Gets the node_ids of this V1ModelServiceRevisionSpec.  # noqa: E501


        :return: The node_ids of this V1ModelServiceRevisionSpec.  # noqa: E501
        :rtype: list[int]
        """
        return self._node_ids

    @node_ids.setter
    def node_ids(self, node_ids):
        """Sets the node_ids of this V1ModelServiceRevisionSpec.


        :param node_ids: The node_ids of this V1ModelServiceRevisionSpec.  # noqa: E501
        :type node_ids: list[int]
        """

        self._node_ids = node_ids

    @property
    def openapi(self):
        """Gets the openapi of this V1ModelServiceRevisionSpec.  # noqa: E501


        :return: The openapi of this V1ModelServiceRevisionSpec.  # noqa: E501
        :rtype: V1OpenAPIForm
        """
        return self._openapi

    @openapi.setter
    def openapi(self, openapi):
        """Sets the openapi of this V1ModelServiceRevisionSpec.


        :param openapi: The openapi of this V1ModelServiceRevisionSpec.  # noqa: E501
        :type openapi: V1OpenAPIForm
        """

        self._openapi = openapi

    @property
    def ports(self):
        """Gets the ports of this V1ModelServiceRevisionSpec.  # noqa: E501


        :return: The ports of this V1ModelServiceRevisionSpec.  # noqa: E501
        :rtype: list[V1Port]
        """
        return self._ports

    @ports.setter
    def ports(self, ports):
        """Sets the ports of this V1ModelServiceRevisionSpec.


        :param ports: The ports of this V1ModelServiceRevisionSpec.  # noqa: E501
        :type ports: list[V1Port]
        """

        self._ports = ports

    @property
    def pre_created_service_account_name(self):
        """Gets the pre_created_service_account_name of this V1ModelServiceRevisionSpec.  # noqa: E501


        :return: The pre_created_service_account_name of this V1ModelServiceRevisionSpec.  # noqa: E501
        :rtype: str
        """
        return self._pre_created_service_account_name

    @pre_created_service_account_name.setter
    def pre_created_service_account_name(self, pre_created_service_account_name):
        """Sets the pre_created_service_account_name of this V1ModelServiceRevisionSpec.


        :param pre_created_service_account_name: The pre_created_service_account_name of this V1ModelServiceRevisionSpec.  # noqa: E501
        :type pre_created_service_account_name: str
        """

        self._pre_created_service_account_name = pre_created_service_account_name

    @property
    def readiness_probe(self):
        """Gets the readiness_probe of this V1ModelServiceRevisionSpec.  # noqa: E501


        :return: The readiness_probe of this V1ModelServiceRevisionSpec.  # noqa: E501
        :rtype: V1ReadinessProbe
        """
        return self._readiness_probe

    @readiness_probe.setter
    def readiness_probe(self, readiness_probe):
        """Sets the readiness_probe of this V1ModelServiceRevisionSpec.


        :param readiness_probe: The readiness_probe of this V1ModelServiceRevisionSpec.  # noqa: E501
        :type readiness_probe: V1ReadinessProbe
        """

        self._readiness_probe = readiness_probe

    @property
    def runner_type(self):
        """Gets the runner_type of this V1ModelServiceRevisionSpec.  # noqa: E501


        :return: The runner_type of this V1ModelServiceRevisionSpec.  # noqa: E501
        :rtype: str
        """
        return self._runner_type

    @runner_type.setter
    def runner_type(self, runner_type):
        """Sets the runner_type of this V1ModelServiceRevisionSpec.


        :param runner_type: The runner_type of this V1ModelServiceRevisionSpec.  # noqa: E501
        :type runner_type: str
        """

        self._runner_type = runner_type

    @property
    def service_account(self):
        """Gets the service_account of this V1ModelServiceRevisionSpec.  # noqa: E501


        :return: The service_account of this V1ModelServiceRevisionSpec.  # noqa: E501
        :rtype: V1ServiceAccount
        """
        return self._service_account

    @service_account.setter
    def service_account(self, service_account):
        """Sets the service_account of this V1ModelServiceRevisionSpec.


        :param service_account: The service_account of this V1ModelServiceRevisionSpec.  # noqa: E501
        :type service_account: V1ServiceAccount
        """

        self._service_account = service_account

    @property
    def volume_claim_mounts(self):
        """Gets the volume_claim_mounts of this V1ModelServiceRevisionSpec.  # noqa: E501


        :return: The volume_claim_mounts of this V1ModelServiceRevisionSpec.  # noqa: E501
        :rtype: dict[str, V1VolumeClaimMount]
        """
        return self._volume_claim_mounts

    @volume_claim_mounts.setter
    def volume_claim_mounts(self, volume_claim_mounts):
        """Sets the volume_claim_mounts of this V1ModelServiceRevisionSpec.


        :param volume_claim_mounts: The volume_claim_mounts of this V1ModelServiceRevisionSpec.  # noqa: E501
        :type volume_claim_mounts: dict[str, V1VolumeClaimMount]
        """

        self._volume_claim_mounts = volume_claim_mounts

    @property
    def volume_v2_mounts(self):
        """Gets the volume_v2_mounts of this V1ModelServiceRevisionSpec.  # noqa: E501


        :return: The volume_v2_mounts of this V1ModelServiceRevisionSpec.  # noqa: E501
        :rtype: list[V1VolumeV2Mount]
        """
        return self._volume_v2_mounts

    @volume_v2_mounts.setter
    def volume_v2_mounts(self, volume_v2_mounts):
        """Sets the volume_v2_mounts of this V1ModelServiceRevisionSpec.


        :param volume_v2_mounts: The volume_v2_mounts of this V1ModelServiceRevisionSpec.  # noqa: E501
        :type volume_v2_mounts: list[V1VolumeV2Mount]
        """

        self._volume_v2_mounts = volume_v2_mounts

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, V1ModelServiceRevisionSpec):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, V1ModelServiceRevisionSpec):
            return True

        return self.to_dict() != other.to_dict()



from .abstract_volume import AbstractVolume
from .file import VolumeFile
from .gcs import GCSVolume
from .s3 import S3Volume
from .storage import StorageType, create_storage, delete_storage, list_storages
from .volume_v2 import (
    create_volume,
    delete_volume,
    delete_volume_file,
    download_volume_file,
    list_volume_files,
    list_volumes,
    upload_volume_file,
    volume_federate,
)

__all__ = [
    "StorageType",
    "create_storage",
    "list_storages",
    "delete_storage",
    "create_volume",
    "list_volumes",
    "delete_volume",
    "list_volume_files",
    "upload_volume_file",
    "download_volume_file",
    "delete_volume_file",
    "volume_federate",
    "AbstractVolume",
    "VolumeFile",
    "GCSVolume",
    "S3Volume",
]
